# -*- coding: utf-8 -*-
"""
print_settings
==============

Django command similar to 'diffsettings' but shows all active Django settings.
"""

import fnmatch
import json

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    help = "Print the active Django settings."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "setting", nargs="*", help="Specifies setting to be printed."
        )
        parser.add_argument(
            "-f",
            "--fail",
            action="store_true",
            dest="fail",
            help="Fail if invalid setting name is given.",
        )
        parser.add_argument(
            "--format", default="simple", dest="format", help="Specifies output format."
        )
        parser.add_argument(
            "--indent",
            default=4,
            dest="indent",
            type=int,
            help="Specifies indent level for JSON and YAML",
        )

    @signalcommand
    def handle(self, *args, **options):
        setting_names = options["setting"]
        settings_dct = {k: getattr(settings, k) for k in dir(settings) if k.isupper()}

        if setting_names:
            settings_dct = {
                key: value
                for key, value in settings_dct.items()
                if any(
                    fnmatch.fnmatchcase(key, setting_name)
                    for setting_name in setting_names
                )
            }

        if options["fail"]:
            for setting_name in setting_names:
                if not any(
                    fnmatch.fnmatchcase(key, setting_name)
                    for key in settings_dct.keys()
                ):
                    raise CommandError("%s not found in settings." % setting_name)

        output_format = options["format"]
        indent = options["indent"]

        if output_format == "json":
            print(json.dumps(settings_dct, indent=indent))
        elif output_format == "yaml":
            import yaml  # requires PyYAML

            print(yaml.dump(settings_dct, indent=indent))
        elif output_format == "pprint":
            from pprint import pprint

            pprint(settings_dct)
        elif output_format == "text":
            for key, value in settings_dct.items():
                print("%s = %s" % (key, value))
        elif output_format == "value":
            for value in settings_dct.values():
                print(value)
        else:
            for key, value in settings_dct.items():
                print("%-40s = %r" % (key, value))



# -*- coding: utf-8 -*-
from django.dispatch import Signal

run_minutely_jobs = Signal()
run_quarter_hourly_jobs = Signal()
run_hourly_jobs = Signal()
run_daily_jobs = Signal()
run_weekly_jobs = Signal()
run_monthly_jobs = Signal()
run_yearly_jobs = Signal()

pre_command = Signal()
post_command = Signal()



# -*- coding: utf-8 -*-
"""
Daily cleanup job.

Can be run as a cronjob to clean out old data from the database (only expired
sessions at the moment).
"""

from django_extensions.management.jobs import DailyJob


class Job(DailyJob):
    help = "Django Daily Cleanup Job"

    def execute(self):
        from django.core import management

        management.call_command("clearsessions")



# -*- coding: utf-8 -*-
import os
from collections import defaultdict

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.db import models

from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    help = "Prints a list of all files in MEDIA_ROOT that are not referenced in the database."  # noqa: E501

    @signalcommand
    def handle(self, *args, **options):
        if not getattr(settings, "MEDIA_ROOT"):
            raise CommandError("MEDIA_ROOT is not set, nothing to do")

        # Get a list of all files under MEDIA_ROOT
        media = set()
        for root, dirs, files in os.walk(settings.MEDIA_ROOT):
            for f in files:
                media.add(os.path.abspath(os.path.join(root, f)))

        # Get list of all fields (value) for each model (key)
        # that is a FileField or subclass of a FileField
        model_dict = defaultdict(list)
        for model in apps.get_models():
            for field in model._meta.fields:
                if issubclass(field.__class__, models.FileField):
                    model_dict[model].append(field)

        # Get a list of all files referenced in the database
        referenced = set()
        for model in model_dict:
            all = model.objects.all().iterator()
            for object in all:
                for field in model_dict[model]:
                    target_file = getattr(object, field.name)
                    if target_file:
                        referenced.add(os.path.abspath(target_file.path))

        # Print each file in MEDIA_ROOT that is not referenced in the database
        not_referenced = media - referenced
        for f in not_referenced:
            print(f)



# -*- coding: utf-8 -*-
import os
import fnmatch

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.core.management.color import color_style
from django.template.loader import get_template

from django_extensions.compat import get_template_setting
from django_extensions.management.utils import signalcommand


#
# TODO: Render the template with fake request object ?
#


class Command(BaseCommand):
    args = ""
    help = "Validate templates on syntax and compile errors"
    ignores = set(
        [
            ".DS_Store",
            "*.swp",
            "*~",
        ]
    )

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--no-apps",
            action="store_true",
            dest="no_apps",
            default=False,
            help="Do not automatically include apps.",
        )
        parser.add_argument(
            "--break",
            "-b",
            action="store_true",
            dest="break",
            default=False,
            help="Break on first error.",
        )
        parser.add_argument(
            "--include",
            "-i",
            action="append",
            dest="includes",
            default=[],
            help="Append these paths to TEMPLATE DIRS",
        )
        parser.add_argument(
            "--ignore-app",
            action="append",
            dest="ignore_apps",
            default=[],
            help="Ignore these apps",
        )

    def ignore_filename(self, filename):
        filename = os.path.basename(filename)
        for ignore_pattern in self.ignores:
            if fnmatch.fnmatch(filename, ignore_pattern):
                return True
        return False

    @signalcommand
    def handle(self, *args, **options):
        if hasattr(settings, "VALIDATE_TEMPLATES_IGNORES"):
            self.ignores = getattr(settings, "VALIDATE_TEMPLATES_IGNORES")

        style = color_style()
        template_dirs = set(get_template_setting("DIRS", []))
        template_dirs |= set(options["includes"])
        template_dirs |= set(
            getattr(settings, "VALIDATE_TEMPLATES_EXTRA_TEMPLATE_DIRS", [])
        )

        if not options["no_apps"]:
            ignore_apps = options["ignore_apps"]
            if not ignore_apps and hasattr(settings, "VALIDATE_TEMPLATES_IGNORE_APPS"):
                ignore_apps = getattr(settings, "VALIDATE_TEMPLATES_IGNORE_APPS")
            for app in apps.get_app_configs():
                if app.name in ignore_apps:
                    continue
                app_template_dir = os.path.join(app.path, "templates")
                if os.path.isdir(app_template_dir):
                    template_dirs.add(app_template_dir)

        settings.TEMPLATES[0]["DIRS"] = list(template_dirs)
        settings.TEMPLATE_DEBUG = True
        verbosity = options["verbosity"]
        errors = 0

        for template_dir in template_dirs:
            for root, dirs, filenames in os.walk(template_dir):
                for filename in filenames:
                    if self.ignore_filename(filename):
                        continue

                    filepath = os.path.join(root, filename)
                    if verbosity > 1:
                        self.stdout.write(filepath)
                    try:
                        get_template(filepath)
                    except Exception as e:
                        errors += 1
                        self.stdout.write(
                            "%s: %s"
                            % (
                                filepath,
                                style.ERROR("%s %s" % (e.__class__.__name__, str(e))),
                            )
                        )
                    if errors and options["break"]:
                        raise CommandError("Errors found")

        if errors:
            raise CommandError("%s errors found" % errors)
        self.stdout.write("%s errors found" % errors)



# -*- coding: utf-8 -*-
from django.core.management import color
from django.utils import termcolors


def _dummy_style_func(msg):
    return msg


def no_style():
    style = color.no_style()
    for role in ("INFO", "WARN", "BOLD", "URL", "MODULE", "MODULE_NAME", "URL_NAME"):
        setattr(style, role, _dummy_style_func)
    return style


def color_style():
    if color.supports_color():
        style = color.color_style()
        style.INFO = termcolors.make_style(fg="green")
        style.WARN = termcolors.make_style(fg="yellow")
        style.BOLD = termcolors.make_style(opts=("bold",))
        style.URL = termcolors.make_style(fg="green", opts=("bold",))
        style.MODULE = termcolors.make_style(fg="yellow")
        style.MODULE_NAME = termcolors.make_style(opts=("bold",))
        style.URL_NAME = termcolors.make_style(fg="red")
    else:
        style = no_style()
    return style



# -*- coding: utf-8 -*-
# Author: AxiaCore S.A.S. https://axiacore.com
from django.conf import settings
from django.core.cache import DEFAULT_CACHE_ALIAS, caches
from django.core.cache.backends.base import InvalidCacheBackendError
from django.core.management.base import BaseCommand, CommandError

from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    """A simple management command which clears the site-wide cache."""

    help = "Fully clear site-wide cache."

    def add_arguments(self, parser):
        parser.add_argument("--cache", action="append", help="Name of cache to clear")
        parser.add_argument(
            "--all",
            "-a",
            action="store_true",
            default=False,
            dest="all_caches",
            help="Clear all configured caches",
        )

    @signalcommand
    def handle(self, cache, all_caches, *args, **kwargs):
        if not cache and not all_caches:
            cache = [DEFAULT_CACHE_ALIAS]
        elif cache and all_caches:
            raise CommandError("Using both --all and --cache is not supported")
        elif all_caches:
            cache = getattr(settings, "CACHES", {DEFAULT_CACHE_ALIAS: {}}).keys()

        for key in cache:
            try:
                caches[key].clear()
            except InvalidCacheBackendError:
                self.stderr.write('Cache "%s" is invalid!\n' % key)
            else:
                self.stdout.write('Cache "%s" has been cleared!\n' % key)



# -*- coding: utf-8 -*-
import sys
import traceback

from django.conf import settings
from django.core.mail import send_mail
from django.core.management import BaseCommand


class EmailNotificationCommand(BaseCommand):
    """
    A BaseCommand subclass which adds sending email fuctionality.

    Subclasses will have an extra command line option ``--email-notification``
    and will be able to send emails by calling ``send_email_notification()``
    if SMTP host and port are specified in settings. The handling of the
    command line option is left to the management command implementation.
    Configuration is done in settings.EMAIL_NOTIFICATIONS dict.

    Configuration example::

        EMAIL_NOTIFICATIONS = {
            'scripts.my_script': {
                'subject': 'my_script subject',
                'body': 'my_script body',
                'from_email': 'from_email@example.com',
                'recipients': ('recipient0@example.com',),
                'no_admins': False,
                'no_traceback': False,
                'notification_level': 0,
                'fail_silently': False
            },
            'scripts.another_script': {
                ...
            },
            ...
        }

    Configuration explained:
        subject:            Email subject.
        body:               Email body.
        from_email:         Email from address.
        recipients:         Sequence of email recipient addresses.
        no_admins:          When True do not include ADMINS to recipients.
        no_traceback:       When True do not include traceback to email body.
        notification_level: 0: send email on fail, 1: send email always.
        fail_silently:      Parameter passed to django's send_mail().
    """

    def add_arguments(self, parser):
        parser.add_argument(
            "--email-notifications",
            action="store_true",
            default=False,
            dest="email_notifications",
            help="Send email notifications for command.",
        )
        parser.add_argument(
            "--email-exception",
            action="store_true",
            default=False,
            dest="email_exception",
            help="Send email for command exceptions.",
        )

    def run_from_argv(self, argv):
        """Overriden in order to access the command line arguments."""
        self.argv_string = " ".join(argv)
        super().run_from_argv(argv)

    def execute(self, *args, **options):
        """
        Overriden in order to send emails on unhandled exception.

        If an unhandled exception in ``def handle(self, *args, **options)``
        occurs and `--email-exception` is set or `self.email_exception` is
        set to True send an email to ADMINS with the traceback and then
        reraise the exception.
        """
        try:
            super().execute(*args, **options)
        except Exception:
            if options["email_exception"] or getattr(self, "email_exception", False):
                self.send_email_notification(include_traceback=True)
            raise

    def send_email_notification(
        self, notification_id=None, include_traceback=False, verbosity=1
    ):
        """
        Send email notifications.

        Reads settings from settings.EMAIL_NOTIFICATIONS dict, if available,
        using ``notification_id`` as a key or else provides reasonable
        defaults.
        """
        # Load email notification settings if available
        if notification_id is not None:
            try:
                email_settings = settings.EMAIL_NOTIFICATIONS.get(notification_id, {})
            except AttributeError:
                email_settings = {}
        else:
            email_settings = {}

        # Exit if no traceback found and not in 'notify always' mode
        if not include_traceback and not email_settings.get("notification_level", 0):
            print(self.style.ERROR("Exiting, not in 'notify always' mode."))
            return

        # Set email fields.
        subject = email_settings.get("subject", "Django extensions email notification.")

        command_name = self.__module__.split(".")[-1]

        body = email_settings.get(
            "body", "Reporting execution of command: '%s'" % command_name
        )

        # Include traceback
        if include_traceback and not email_settings.get("no_traceback", False):
            try:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                trb = "".join(traceback.format_tb(exc_traceback))
                body += "\n\nTraceback:\n\n%s\n" % trb
            finally:
                del exc_traceback

        # Set from address
        from_email = email_settings.get("from_email", settings.DEFAULT_FROM_EMAIL)

        # Calculate recipients
        recipients = list(email_settings.get("recipients", []))

        if not email_settings.get("no_admins", False):
            recipients.extend(settings.ADMINS)

        if not recipients:
            if verbosity > 0:
                print(self.style.ERROR("No email recipients available."))
            return

        # Send email...
        send_mail(
            subject,
            body,
            from_email,
            recipients,
            fail_silently=email_settings.get("fail_silently", True),
        )



# -*- coding: utf-8 -*-
import inspect
import os
import sys
import traceback
import warnings

from django.db import connections
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.utils.datastructures import OrderedSet

from django_extensions.management.shells import import_objects
from django_extensions.management.utils import signalcommand
from django_extensions.management.debug_cursor import monkey_patch_cursordebugwrapper


def use_vi_mode():
    editor = os.environ.get("EDITOR")
    if not editor:
        return False
    editor = os.path.basename(editor)
    return editor.startswith("vi") or editor.endswith("vim")


def shell_runner(flags, name, help=None):
    """
    Decorates methods with information about the application they are starting

    :param flags: The flags used to start this runner via the ArgumentParser.
    :param name: The name of this runner for the help text for the ArgumentParser.
    :param help: The optional help for the ArgumentParser if the dynamically generated help is not sufficient.
    """  # noqa: E501

    def decorator(fn):
        fn.runner_flags = flags
        fn.runner_name = name
        fn.runner_help = help

        return fn

    return decorator


class Command(BaseCommand):
    help = "Like the 'shell' command but autoloads the models of all installed Django apps."  # noqa: E501
    extra_args = None
    tests_mode = False

    def __init__(self):
        super().__init__()
        self.runners = [
            member
            for name, member in inspect.getmembers(self)
            if hasattr(member, "runner_flags")
        ]

    def add_arguments(self, parser):
        super().add_arguments(parser)

        group = parser.add_mutually_exclusive_group()
        for runner in self.runners:
            if runner.runner_help:
                help = runner.runner_help
            else:
                help = "Tells Django to use %s." % runner.runner_name

            group.add_argument(
                *runner.runner_flags,
                action="store_const",
                dest="runner",
                const=runner,
                help=help,
            )

        parser.add_argument(
            "--connection-file",
            action="store",
            dest="connection_file",
            help="Specifies the connection file to use if using the --kernel option",
        )
        parser.add_argument(
            "--no-startup",
            action="store_true",
            dest="no_startup",
            default=False,
            help=(
                "When using plain Python, ignore the PYTHONSTARTUP environment "
                "variable and ~/.pythonrc.py script."
            ),
        )
        parser.add_argument(
            "--use-pythonrc",
            action="store_true",
            dest="use_pythonrc",
            default=False,
            help=(
                "When using plain Python, load the PYTHONSTARTUP environment variable "
                "and ~/.pythonrc.py script."
            ),
        )
        parser.add_argument(
            "--print-sql",
            action="store_true",
            default=False,
            help="Print SQL queries as they're executed",
        )
        parser.add_argument(
            "--truncate-sql",
            action="store",
            type=int,
            help="Truncate SQL queries to a number of characters.",
        )
        parser.add_argument(
            "--print-sql-location",
            action="store_true",
            default=False,
            help="Show location in code where SQL query generated from",
        )
        parser.add_argument(
            "--dont-load",
            action="append",
            dest="dont_load",
            default=[],
            help="Ignore autoloading of some apps/models. Can be used several times.",
        )
        parser.add_argument(
            "--quiet-load",
            action="store_true",
            default=False,
            dest="quiet_load",
            help="Do not display loaded models messages",
        )
        parser.add_argument(
            "--vi",
            action="store_true",
            default=use_vi_mode(),
            dest="vi_mode",
            help="Load Vi key bindings (for --ptpython and --ptipython)",
        )
        parser.add_argument(
            "--no-browser",
            action="store_true",
            default=False,
            dest="no_browser",
            help="Don't open the notebook in a browser after startup.",
        )
        parser.add_argument(
            "-c",
            "--command",
            help=(
                "Instead of opening an interactive shell, "
                "run a command as Django and exit."
            ),
        )

    def run_from_argv(self, argv):
        if "--" in argv[2:]:
            idx = argv.index("--")
            self.extra_args = argv[idx + 1 :]
            argv = argv[:idx]
        return super().run_from_argv(argv)

    def get_ipython_arguments(self, options):
        ipython_args = "IPYTHON_ARGUMENTS"
        arguments = getattr(settings, ipython_args, [])
        if not arguments:
            arguments = os.environ.get(ipython_args, "").split()
        return arguments

    def get_notebook_arguments(self, options):
        notebook_args = "NOTEBOOK_ARGUMENTS"
        arguments = getattr(settings, notebook_args, [])
        if not arguments:
            arguments = os.environ.get(notebook_args, "").split()
        return arguments

    def get_imported_objects(self, options):
        imported_objects = import_objects(options, self.style)
        if self.tests_mode:
            # save imported objects so we can run tests against it later
            self.tests_imported_objects = imported_objects
        return imported_objects

    @shell_runner(flags=["--kernel"], name="IPython Kernel")
    def get_kernel(self, options):
        try:
            from IPython import release

            if release.version_info[0] < 2:
                print(
                    self.style.ERROR("--kernel requires at least IPython version 2.0")
                )
                return
            from IPython import start_kernel
        except ImportError:
            return traceback.format_exc()

        def run_kernel():
            imported_objects = self.get_imported_objects(options)
            kwargs = dict(
                argv=[],
                user_ns=imported_objects,
            )
            connection_file = options["connection_file"]
            if connection_file:
                kwargs["connection_file"] = connection_file
            start_kernel(**kwargs)

        return run_kernel

    def load_base_kernel_spec(self, app):
        """Finds and returns the base Python kernelspec to extend from."""
        ksm = app.kernel_spec_manager
        try_spec_names = getattr(
            settings,
            "NOTEBOOK_KERNEL_SPEC_NAMES",
            [
                "python3",
                "python",
            ],
        )

        if isinstance(try_spec_names, str):
            try_spec_names = [try_spec_names]

        ks = None
        for spec_name in try_spec_names:
            try:
                ks = ksm.get_kernel_spec(spec_name)
                break
            except Exception:
                continue
        if not ks:
            raise CommandError(
                "No notebook (Python) kernel specs found. Tried %r" % try_spec_names
            )

        return ks

    def generate_kernel_specs(self, app, ipython_arguments):
        """Generate an IPython >= 3.0 kernelspec that loads django extensions"""
        ks = self.load_base_kernel_spec(app)
        ks.argv.extend(ipython_arguments)
        ks.display_name = getattr(
            settings, "IPYTHON_KERNEL_DISPLAY_NAME", "Django Shell-Plus"
        )

        manage_py_dir, manage_py = os.path.split(os.path.realpath(sys.argv[0]))
        if manage_py == "manage.py" and os.path.isdir(manage_py_dir):
            pythonpath = ks.env.get("PYTHONPATH", os.environ.get("PYTHONPATH", ""))
            pythonpath = pythonpath.split(os.pathsep)
            if manage_py_dir not in pythonpath:
                pythonpath.append(manage_py_dir)

            ks.env["PYTHONPATH"] = os.pathsep.join(filter(None, pythonpath))

        return {"django_extensions": ks}

    def run_notebookapp(self, app_init, options, use_kernel_specs=True, history=True):
        no_browser = options["no_browser"]

        if self.extra_args:
            # if another '--' is found split the arguments notebook, ipython
            if "--" in self.extra_args:
                idx = self.extra_args.index("--")
                notebook_arguments = self.extra_args[:idx]
                ipython_arguments = self.extra_args[idx + 1 :]
            # otherwise pass the arguments to the notebook
            else:
                notebook_arguments = self.extra_args
                ipython_arguments = []
        else:
            notebook_arguments = self.get_notebook_arguments(options)
            ipython_arguments = self.get_ipython_arguments(options)

        # Treat IPYTHON_ARGUMENTS from settings
        if "django_extensions.management.notebook_extension" not in ipython_arguments:
            ipython_arguments.extend(
                ["--ext", "django_extensions.management.notebook_extension"]
            )

        # Treat NOTEBOOK_ARGUMENTS from settings
        if no_browser and "--no-browser" not in notebook_arguments:
            notebook_arguments.append("--no-browser")
        if "--notebook-dir" not in notebook_arguments and not any(
            e.startswith("--notebook-dir=") for e in notebook_arguments
        ):
            notebook_arguments.extend(["--notebook-dir", "."])

        # IPython < 3 passes through kernel args from notebook CLI
        if not use_kernel_specs:
            notebook_arguments.extend(ipython_arguments)

        # disable history if not already configured in some other way
        if not history and not any(
            arg.startswith("--HistoryManager") for arg in ipython_arguments
        ):
            ipython_arguments.append("--HistoryManager.enabled=False")

        if not callable(app_init):
            app = app_init
            warnings.warn(
                "Initialize should be a callable not an app instance",
                DeprecationWarning,
            )
            app.initialize(notebook_arguments)
        else:
            app = app_init(notebook_arguments)

        # IPython >= 3 uses kernelspecs to specify kernel CLI args
        if use_kernel_specs:
            ksm = app.kernel_spec_manager
            for kid, ks in self.generate_kernel_specs(app, ipython_arguments).items():
                roots = [os.path.dirname(ks.resource_dir), ksm.user_kernel_dir]

                for root in roots:
                    kernel_dir = os.path.join(root, kid)
                    try:
                        if not os.path.exists(kernel_dir):
                            os.makedirs(kernel_dir)
                        with open(os.path.join(kernel_dir, "kernel.json"), "w") as f:
                            f.write(ks.to_json())
                        break
                    except OSError:
                        continue
                else:
                    raise CommandError(
                        "Could not write kernel %r in directories %r" % (kid, roots)
                    )

        app.start()

    @shell_runner(flags=["--notebook"], name="IPython Notebook")
    def get_notebook(self, options):
        try:
            from IPython import release
        except ImportError:
            return traceback.format_exc()
        try:
            from notebook.notebookapp import NotebookApp
        except ImportError:
            if release.version_info[0] >= 7:
                return traceback.format_exc()
            try:
                from IPython.html.notebookapp import NotebookApp
            except ImportError:
                if release.version_info[0] >= 3:
                    return traceback.format_exc()
                try:
                    from IPython.frontend.html.notebook import notebookapp

                    NotebookApp = notebookapp.NotebookApp
                except ImportError:
                    return traceback.format_exc()

        use_kernel_specs = release.version_info[0] >= 3

        def app_init(*args, **kwargs):
            app = NotebookApp.instance()
            app.initialize(*args, **kwargs)
            return app

        def run_notebook():
            self.run_notebookapp(app_init, options, use_kernel_specs)

        return run_notebook

    @shell_runner(flags=["--lab"], name="JupyterLab Notebook")
    def get_jupyterlab(self, options):
        try:
            from jupyterlab.labapp import LabApp
        except ImportError:
            return traceback.format_exc()

        # check for JupyterLab 3.0
        try:
            from notebook.notebookapp import NotebookApp
        except ImportError:
            NotebookApp = None

        if not NotebookApp or not issubclass(LabApp, NotebookApp):
            app_init = LabApp.initialize_server
        else:

            def app_init(*args, **kwargs):
                app = LabApp.instance()
                app.initialize(*args, **kwargs)
                return app

        def run_jupyterlab():
            self.run_notebookapp(app_init, options, history=False)

        return run_jupyterlab

    @shell_runner(flags=["--plain"], name="plain Python")
    def get_plain(self, options):
        # Using normal Python shell
        import code

        # Set up a dictionary to serve as the environment for the shell.
        imported_objects = self.get_imported_objects(options)

        use_pythonrc = options["use_pythonrc"]
        no_startup = options["no_startup"]

        # We want to honor both $PYTHONSTARTUP and .pythonrc.py, so follow system
        # conventions and get $PYTHONSTARTUP first then .pythonrc.py.
        if use_pythonrc or not no_startup:
            for pythonrc in OrderedSet(
                [os.environ.get("PYTHONSTARTUP"), os.path.expanduser("~/.pythonrc.py")]
            ):
                if not pythonrc:
                    continue
                if not os.path.isfile(pythonrc):
                    continue
                with open(pythonrc) as handle:
                    pythonrc_code = handle.read()
                # Match the behavior of the cpython shell where an error in
                # PYTHONSTARTUP prints an exception and continues.
                try:
                    exec(compile(pythonrc_code, pythonrc, "exec"), imported_objects)
                except Exception:
                    traceback.print_exc()
                    if self.tests_mode:
                        raise

        # By default, this will set up readline to do tab completion and to read and
        # write history to the .python_history file, but this can be overridden by
        # $PYTHONSTARTUP or ~/.pythonrc.py.
        try:
            hook = sys.__interactivehook__
        except AttributeError:
            # Match the behavior of the cpython shell where a missing
            # sys.__interactivehook__ is ignored.
            pass
        else:
            try:
                hook()
            except Exception:
                # Match the behavior of the cpython shell where an error in
                # sys.__interactivehook__ prints a warning and the exception
                # and continues.
                print("Failed calling sys.__interactivehook__")
                traceback.print_exc()

        try:
            # Try activating rlcompleter, because it's handy.
            import readline
        except ImportError:
            pass
        else:
            # We don't have to wrap the following import in a 'try', because
            # we already know 'readline' was imported successfully.
            import rlcompleter

            readline.set_completer(rlcompleter.Completer(imported_objects).complete)
            # Enable tab completion on systems using libedit (e.g. macOS).
            # These lines are copied from Lib/site.py on Python 3.4.
            readline_doc = getattr(readline, "__doc__", "")
            if readline_doc is not None and "libedit" in readline_doc:
                readline.parse_and_bind("bind ^I rl_complete")
            else:
                readline.parse_and_bind("tab:complete")

        def run_plain():
            code.interact(local=imported_objects)

        return run_plain

    @shell_runner(flags=["--bpython"], name="BPython")
    def get_bpython(self, options):
        try:
            from bpython import embed
        except ImportError:
            return traceback.format_exc()

        def run_bpython():
            imported_objects = self.get_imported_objects(options)
            kwargs = {}
            if self.extra_args:
                kwargs["args"] = self.extra_args
            embed(imported_objects, **kwargs)

        return run_bpython

    @shell_runner(flags=["--ipython"], name="IPython")
    def get_ipython(self, options):
        try:
            from IPython import start_ipython

            def run_ipython():
                imported_objects = self.get_imported_objects(options)
                ipython_arguments = self.extra_args or self.get_ipython_arguments(
                    options
                )
                start_ipython(argv=ipython_arguments, user_ns=imported_objects)

            return run_ipython
        except ImportError:
            str_exc = traceback.format_exc()
            # IPython < 0.11
            # Explicitly pass an empty list as arguments, because otherwise
            # IPython would use sys.argv from this script.
            # Notebook not supported for IPython < 0.11.
            try:
                from IPython.Shell import IPShell
            except ImportError:
                return str_exc + "\n" + traceback.format_exc()

            def run_ipython():
                imported_objects = self.get_imported_objects(options)
                shell = IPShell(argv=[], user_ns=imported_objects)
                shell.mainloop()

            return run_ipython

    @shell_runner(flags=["--ptpython"], name="PTPython")
    def get_ptpython(self, options):
        try:
            from ptpython.repl import embed, run_config
        except ImportError:
            tb = traceback.format_exc()
            try:  # prompt_toolkit < v0.27
                from prompt_toolkit.contrib.repl import embed, run_config
            except ImportError:
                return tb

        def run_ptpython():
            imported_objects = self.get_imported_objects(options)
            history_filename = os.path.expanduser("~/.ptpython_history")
            embed(
                globals=imported_objects,
                history_filename=history_filename,
                vi_mode=options["vi_mode"],
                configure=run_config,
            )

        return run_ptpython

    @shell_runner(flags=["--ptipython"], name="PT-IPython")
    def get_ptipython(self, options):
        try:
            from ptpython.repl import run_config
            from ptpython.ipython import embed
        except ImportError:
            tb = traceback.format_exc()
            try:  # prompt_toolkit < v0.27
                from prompt_toolkit.contrib.repl import run_config
                from prompt_toolkit.contrib.ipython import embed
            except ImportError:
                return tb

        def run_ptipython():
            imported_objects = self.get_imported_objects(options)
            history_filename = os.path.expanduser("~/.ptpython_history")
            embed(
                user_ns=imported_objects,
                history_filename=history_filename,
                vi_mode=options["vi_mode"],
                configure=run_config,
            )

        return run_ptipython

    @shell_runner(flags=["--idle"], name="Idle")
    def get_idle(self, options):
        from idlelib.pyshell import main

        def run_idle():
            sys.argv = [
                sys.argv[0],
                "-c",
                """
from django_extensions.management import shells
from django.core.management.color import no_style
for k, m in shells.import_objects({}, no_style()).items():
    globals()[k] = m
""",
            ]
            main()

        return run_idle

    def set_application_name(self, options):
        """
        Set the application_name on PostgreSQL connection

        Use the fallback_application_name to let the user override
        it with PGAPPNAME env variable

        https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS
        """  # noqa: E501
        supported_backends = (
            "django.db.backends.postgresql",
            "django.db.backends.postgresql_psycopg2",
        )
        opt_name = "fallback_application_name"
        default_app_name = "django_shell"
        dbs = getattr(settings, "DATABASES", [])

        for connection in connections.all():
            alias = connection.alias
            mro = inspect.getmro(connection.__class__)
            if any(klass.__module__.startswith(supported_backends) for klass in mro):
                if "OPTIONS" not in dbs[alias] or opt_name not in dbs[alias]["OPTIONS"]:
                    dbs[alias].setdefault("OPTIONS", {}).update(
                        {opt_name: default_app_name}
                    )

    @signalcommand
    def handle(self, *args, **options):
        verbosity = options["verbosity"]
        get_runner = options["runner"]
        print_sql = getattr(settings, "SHELL_PLUS_PRINT_SQL", False)
        runner = None
        runner_name = None
        truncate = None if options["truncate_sql"] == 0 else options["truncate_sql"]

        with monkey_patch_cursordebugwrapper(
            print_sql=options["print_sql"] or print_sql,
            truncate=truncate,
            print_sql_location=options["print_sql_location"],
            confprefix="SHELL_PLUS",
        ):
            SETTINGS_SHELL_PLUS = getattr(settings, "SHELL_PLUS", None)

            def get_runner_by_flag(flag):
                for runner in self.runners:
                    if flag in runner.runner_flags:
                        return runner
                return None

            self.set_application_name(options)

            if not get_runner and SETTINGS_SHELL_PLUS:
                get_runner = get_runner_by_flag("--%s" % SETTINGS_SHELL_PLUS)
                if not get_runner:
                    runner = None
                    runner_name = SETTINGS_SHELL_PLUS

            if get_runner:
                runner = get_runner(options)
                runner_name = get_runner.runner_name
            else:

                def try_runner(get_runner):
                    runner_name = get_runner.runner_name
                    if verbosity > 2:
                        print(self.style.NOTICE("Trying: %s" % runner_name))

                    runner = get_runner(options)
                    if callable(runner):
                        if verbosity > 1:
                            print(self.style.NOTICE("Using: %s" % runner_name))
                        return runner
                    return None

                tried_runners = set()

                # try the runners that are least unexpected (normal shell runners)
                preferred_runners = [
                    "ptipython",
                    "ptpython",
                    "bpython",
                    "ipython",
                    "plain",
                ]
                for flag_suffix in preferred_runners:
                    get_runner = get_runner_by_flag("--%s" % flag_suffix)
                    tried_runners.add(get_runner)
                    runner = try_runner(get_runner)
                    if runner:
                        runner_name = get_runner.runner_name
                        break

                # try any remaining runners if needed
                if not runner:
                    for get_runner in self.runners:
                        if get_runner not in tried_runners:
                            runner = try_runner(get_runner)
                            if runner:
                                runner_name = get_runner.runner_name
                                break

            if not callable(runner):
                if runner:
                    print(runner)
                if not runner_name:
                    raise CommandError("No shell runner could be found.")
                raise CommandError("Could not load shell runner: '%s'." % runner_name)

            if self.tests_mode:
                return 130

            if options["command"]:
                imported_objects = self.get_imported_objects(options)
                exec(options["command"], imported_objects)
                return None

            runner()



# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand

from django_extensions.management.utils import signalcommand


class DjangoExtensionsTestException(Exception):
    pass


class Command(BaseCommand):
    help = (
        "Raises a test Exception named DjangoExtensionsTestException. "
        "Useful for debugging integration with error reporters like Sentry."
    )

    @signalcommand
    def handle(self, *args, **options):
        message = (
            "This is a test exception via the "
            "django-extensions raise_test_exception management command."
        )
        raise DjangoExtensionsTestException(message)



# -*- coding: utf-8 -*-
import asyncio
import sys

try:
    from aiosmtpd.controller import Controller
except ImportError:
    raise ImportError("Please install 'aiosmtpd' library to use mail_debug command.")

from logging import getLogger
from typing import List

from django.core.management.base import BaseCommand, CommandError

from django_extensions.management.utils import setup_logger, signalcommand

logger = getLogger(__name__)


class CustomHandler:
    async def handle_DATA(self, server, session, envelope):
        """Output will be sent to the module logger at INFO level."""
        peer = session.peer
        inheaders = 1
        lines = envelope.content.decode("utf8", errors="replace").splitlines()
        logger.info("---------- MESSAGE FOLLOWS ----------")
        for line in lines:
            # headers first
            if inheaders and not line:
                logger.info("X-Peer: %s" % peer[0])
                inheaders = 0
            logger.info(line)
        logger.info("------------ END MESSAGE ------------")
        return "250 OK"


class Command(BaseCommand):
    help = "Starts a test mail server for development."
    args = "[optional port number or ippaddr:port]"

    requires_system_checks: List[str] = []

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("addrport", nargs="?")
        parser.add_argument(
            "--output",
            dest="output_file",
            default=None,
            help="Specifies an output file to send a copy of all messages "
            "(not flushed immediately).",
        )
        parser.add_argument(
            "--use-settings",
            dest="use_settings",
            action="store_true",
            default=False,
            help="Uses EMAIL_HOST and HOST_PORT from Django settings.",
        )

    @signalcommand
    def handle(self, addrport="", *args, **options):
        if not addrport:
            if options["use_settings"]:
                from django.conf import settings

                addr = getattr(settings, "EMAIL_HOST", "")
                port = str(getattr(settings, "EMAIL_PORT", "1025"))
            else:
                addr = ""
                port = "1025"
        else:
            try:
                addr, port = addrport.split(":")
            except ValueError:
                addr, port = "", addrport
        if not addr:
            addr = "127.0.0.1"

        if not port.isdigit():
            raise CommandError("%r is not a valid port number." % port)
        else:
            port = int(port)

        # Add console handler
        setup_logger(logger, stream=self.stdout, filename=options["output_file"])

        def inner_run():
            quit_command = (sys.platform == "win32") and "CTRL-BREAK" or "CONTROL-C"
            print(
                "Now accepting mail at %s:%s -- use %s to quit"
                % (addr, port, quit_command)
            )
            handler = CustomHandler()
            controller = Controller(handler, hostname=addr, port=port)
            controller.start()
            loop = asyncio.get_event_loop()
            loop.run_forever()

        try:
            inner_run()
        except KeyboardInterrupt:
            pass



# -*- coding: utf-8 -*-
import os

from django.conf import settings

BASE_DIR = os.path.dirname(os.path.realpath(__file__))
REPLACEMENTS = getattr(settings, "EXTENSIONS_REPLACEMENTS", {})

DEFAULT_SQLITE_ENGINES = (
    "django.db.backends.sqlite3",
    "django.db.backends.spatialite",
    "django_prometheus.db.backends.sqlite3",
)
DEFAULT_MYSQL_ENGINES = (
    "django.db.backends.mysql",
    "django.contrib.gis.db.backends.mysql",
    "django_prometheus.db.backends.mysql",
    "mysql.connector.django",
)
DEFAULT_POSTGRESQL_ENGINES = (
    "django.db.backends.postgresql",
    "django.db.backends.postgresql_psycopg2",
    "django.db.backends.postgis",
    "django.contrib.gis.db.backends.postgis",
    "psqlextra.backend",
    "django_zero_downtime_migrations.backends.postgres",
    "django_zero_downtime_migrations.backends.postgis",
    "django_prometheus.db.backends.postgresql",
    "django_prometheus.db.backends.postgis",
    "django_tenants.postgresql_backend",
)

SQLITE_ENGINES = getattr(
    settings, "DJANGO_EXTENSIONS_RESET_DB_SQLITE_ENGINES", DEFAULT_SQLITE_ENGINES
)
MYSQL_ENGINES = getattr(
    settings, "DJANGO_EXTENSIONS_RESET_DB_MYSQL_ENGINES", DEFAULT_MYSQL_ENGINES
)
POSTGRESQL_ENGINES = getattr(
    settings,
    "DJANGO_EXTENSIONS_RESET_DB_POSTGRESQL_ENGINES",
    DEFAULT_POSTGRESQL_ENGINES,
)

DEFAULT_PRINT_SQL_TRUNCATE_CHARS = 1000



# -*- coding: utf-8 -*-
from typing import List

from django.core.management.base import BaseCommand
from django.core.management.utils import get_random_secret_key

from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    help = "Generates a new SECRET_KEY that can be used in a project settings file."

    requires_system_checks: List[str] = []

    @signalcommand
    def handle(self, *args, **options):
        return get_random_secret_key()



# -*- coding: utf-8 -*-
def load_ipython_extension(ipython):
    from django.core.management.color import no_style
    from django_extensions.management.shells import import_objects

    imported_objects = import_objects(
        options={"dont_load": []},
        style=no_style(),
    )
    ipython.push(imported_objects)



# -*- coding: utf-8 -*-
import fnmatch
import os
from os.path import join as _j
from typing import List

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    help = "Removes all python bytecode compiled files from the project."

    requires_system_checks: List[str] = []

    def add_arguments(self, parser):
        parser.add_argument(
            "--optimize",
            "-o",
            "-O",
            action="store_true",
            dest="optimize",
            default=False,
            help="Remove optimized python bytecode files",
        )
        parser.add_argument(
            "--path",
            "-p",
            action="store",
            dest="path",
            help="Specify path to recurse into",
        )

    @signalcommand
    def handle(self, *args, **options):
        project_root = options.get("path", getattr(settings, "BASE_DIR", None))
        if not project_root:
            project_root = getattr(settings, "BASE_DIR", None)

        verbosity = options["verbosity"]
        if not project_root:
            raise CommandError(
                "No --path specified and settings.py does not contain BASE_DIR"
            )

        exts = options["optimize"] and "*.py[co]" or "*.pyc"

        for root, dirs, filenames in os.walk(project_root):
            for filename in fnmatch.filter(filenames, exts):
                full_path = _j(root, filename)
                if verbosity > 1:
                    self.stdout.write("%s\n" % full_path)
                os.remove(full_path)



# -*- coding: utf-8 -*-
from django.contrib.auth.mixins import UserPassesTestMixin


class ModelUserFieldPermissionMixin(UserPassesTestMixin):
    model_permission_user_field = "user"

    def get_model_permission_user_field(self):
        return self.model_permission_user_field

    def test_func(self):
        model_attr = self.get_model_permission_user_field()
        current_user = self.request.user

        return current_user == getattr(self.get_queryset().first(), model_attr)



# -*- coding: utf-8 -*-
import sys
import json
import os
import tempfile

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.template import loader

from django_extensions.management.modelviz import ModelGraph, generate_dot
from django_extensions.management.utils import signalcommand

try:
    import pygraphviz

    HAS_PYGRAPHVIZ = True
except ImportError:
    HAS_PYGRAPHVIZ = False

try:
    try:
        import pydotplus as pydot
    except ImportError:
        import pydot
    HAS_PYDOT = True
except ImportError:
    HAS_PYDOT = False


def retheme(graph_data, app_style={}):
    if isinstance(app_style, str):
        if os.path.exists(app_style):
            try:
                with open(app_style, "rt") as f:
                    app_style = json.load(f)
            except Exception as e:
                print(f"Invalid app style file {app_style}")
                raise Exception(e)
        else:
            return graph_data

    for gc in graph_data["graphs"]:
        for g in gc:
            if "name" in g:
                for m in g["models"]:
                    app_name = g["app_name"]
                    if app_name in app_style:
                        m["style"] = app_style[app_name]
    return graph_data


class Command(BaseCommand):
    help = "Creates a GraphViz dot file for the specified app names."
    " You can pass multiple app names and they will all be combined into a"
    " single model. Output is usually directed to a dot file."

    can_import_settings = True

    def __init__(self, *args, **kwargs):
        """
        Allow defaults for arguments to be set in settings.GRAPH_MODELS.

        Each argument in self.arguments is a dict where the key is the
        space-separated args and the value is our kwarg dict.

        The default from settings is keyed as the long arg name with '--'
        removed and any '-' replaced by '_'. For example, the default value for
        --disable-fields can be set in settings.GRAPH_MODELS['disable_fields'].
        """
        self.arguments = {
            "--app-style": {
                "action": "store",
                "help": "Path to style json to configure the style per app",
                "dest": "app-style",
                "default": ".app-style.json",
            },
            "--pygraphviz": {
                "action": "store_true",
                "default": False,
                "dest": "pygraphviz",
                "help": "Output graph data as image using PyGraphViz.",
            },
            "--pydot": {
                "action": "store_true",
                "default": False,
                "dest": "pydot",
                "help": "Output graph data as image using PyDot(Plus).",
            },
            "--dot": {
                "action": "store_true",
                "default": False,
                "dest": "dot",
                "help": (
                    "Output graph data as raw DOT (graph description language) "
                    "text data."
                ),
            },
            "--json": {
                "action": "store_true",
                "default": False,
                "dest": "json",
                "help": "Output graph data as JSON",
            },
            "--disable-fields -d": {
                "action": "store_true",
                "default": False,
                "dest": "disable_fields",
                "help": "Do not show the class member fields",
            },
            "--disable-abstract-fields": {
                "action": "store_true",
                "default": False,
                "dest": "disable_abstract_fields",
                "help": "Do not show the class member fields that were inherited",
            },
            "--display-field-choices": {
                "action": "store_true",
                "default": False,
                "dest": "display_field_choices",
                "help": "Display choices instead of field type",
            },
            "--group-models -g": {
                "action": "store_true",
                "default": False,
                "dest": "group_models",
                "help": "Group models together respective to their application",
            },
            "--all-applications -a": {
                "action": "store_true",
                "default": False,
                "dest": "all_applications",
                "help": "Automatically include all applications from INSTALLED_APPS",
            },
            "--output -o": {
                "action": "store",
                "dest": "outputfile",
                "help": (
                    "Render output file. Type of output dependend on file extensions. "
                    "Use png or jpg to render graph to image."
                ),
            },
            "--layout -l": {
                "action": "store",
                "dest": "layout",
                "default": "dot",
                "help": "Layout to be used by GraphViz for visualization. Layouts: "
                "circo dot fdp neato nop nop1 nop2 twopi",
            },
            "--theme -t": {
                "action": "store",
                "dest": "theme",
                "default": "django2018",
                "help": "Theme to use. Supplied are 'original' and 'django2018'. "
                "You can create your own by creating dot templates in "
                "'django_extentions/graph_models/themename/' template directory.",
            },
            "--verbose-names -n": {
                "action": "store_true",
                "default": False,
                "dest": "verbose_names",
                "help": "Use verbose_name of models and fields",
            },
            "--language -L": {
                "action": "store",
                "dest": "language",
                "help": "Specify language used for verbose_name localization",
            },
            "--exclude-columns -x": {
                "action": "store",
                "dest": "exclude_columns",
                "help": "Exclude specific column(s) from the graph. "
                "Can also load exclude list from file.",
            },
            "--exclude-models -X": {
                "action": "store",
                "dest": "exclude_models",
                "help": "Exclude specific model(s) from the graph. Can also load "
                "exclude list from file. Wildcards (*) are allowed.",
            },
            "--include-models -I": {
                "action": "store",
                "dest": "include_models",
                "help": "Restrict the graph to specified models. "
                "Wildcards (*) are allowed.",
            },
            "--inheritance -e": {
                "action": "store_true",
                "default": True,
                "dest": "inheritance",
                "help": "Include inheritance arrows (default)",
            },
            "--no-inheritance -E": {
                "action": "store_false",
                "default": False,
                "dest": "inheritance",
                "help": "Do not include inheritance arrows",
            },
            "--hide-relations-from-fields -R": {
                "action": "store_false",
                "default": True,
                "dest": "relations_as_fields",
                "help": "Do not show relations as fields in the graph.",
            },
            "--relation-fields-only": {
                "action": "store",
                "default": False,
                "dest": "relation_fields_only",
                "help": "Only display fields that are relevant for relations",
            },
            "--disable-sort-fields -S": {
                "action": "store_false",
                "default": True,
                "dest": "sort_fields",
                "help": "Do not sort fields",
            },
            "--hide-edge-labels": {
                "action": "store_true",
                "default": False,
                "dest": "hide_edge_labels",
                "help": "Do not show relations labels in the graph.",
            },
            "--arrow-shape": {
                "action": "store",
                "default": "dot",
                "dest": "arrow_shape",
                "choices": [
                    "box",
                    "crow",
                    "curve",
                    "icurve",
                    "diamond",
                    "dot",
                    "inv",
                    "none",
                    "normal",
                    "tee",
                    "vee",
                ],
                "help": "Arrow shape to use for relations. Default is dot. "
                "Available shapes: box, crow, curve, icurve, diamond, dot, inv, "
                "none, normal, tee, vee.",
            },
            "--color-code-deletions": {
                "action": "store_true",
                "default": False,
                "dest": "color_code_deletions",
                "help": "Color the relations according to their on_delete setting, "
                "where it is applicable. The colors are: red (CASCADE), "
                "orange (SET_NULL), green (SET_DEFAULT), yellow (SET), "
                "blue (PROTECT), grey (DO_NOTHING), and purple (RESTRICT).",
            },
            "--rankdir": {
                "action": "store",
                "default": "TB",
                "dest": "rankdir",
                "choices": ["TB", "BT", "LR", "RL"],
                "help": "Set direction of graph layout. Supported directions: "
                "TB, LR, BT and RL. Corresponding to directed graphs drawn from "
                "top to bottom, from left to right, from bottom to top, and from "
                "right to left, respectively. Default is TB.",
            },
            "--ordering": {
                "action": "store",
                "default": None,
                "dest": "ordering",
                "choices": ["in", "out"],
                "help": "Controls how the edges are arranged. Supported orderings: "
                '"in" (incoming relations first), "out" (outgoing relations first). '
                "Default is None.",
            },
        }

        defaults = getattr(settings, "GRAPH_MODELS", None)

        if defaults:
            for argument in self.arguments:
                arg_split = argument.split(" ")
                setting_opt = arg_split[0].lstrip("-").replace("-", "_")
                if setting_opt in defaults:
                    self.arguments[argument]["default"] = defaults[setting_opt]

        super().__init__(*args, **kwargs)

    def add_arguments(self, parser):
        """Unpack self.arguments for parser.add_arguments."""
        parser.add_argument("app_label", nargs="*")
        for argument in self.arguments:
            parser.add_argument(*argument.split(" "), **self.arguments[argument])

    @signalcommand
    def handle(self, *args, **options):
        args = options["app_label"]
        if not args and not options["all_applications"]:
            default_app_labels = getattr(settings, "GRAPH_MODELS", {}).get("app_labels")
            if default_app_labels:
                args = default_app_labels
            else:
                raise CommandError("need one or more arguments for appname")

        # Determine output format based on options, file extension, and library
        # availability.
        outputfile = options.get("outputfile") or ""
        _, outputfile_ext = os.path.splitext(outputfile)
        outputfile_ext = outputfile_ext.lower()
        output_opts_names = ["pydot", "pygraphviz", "json", "dot"]
        output_opts = {k: v for k, v in options.items() if k in output_opts_names}
        output_opts_count = sum(output_opts.values())
        if output_opts_count > 1:
            raise CommandError(
                "Only one of %s can be set."
                % ", ".join(["--%s" % opt for opt in output_opts_names])
            )

        if output_opts_count == 1:
            output = next(key for key, val in output_opts.items() if val)
        elif not outputfile:
            # When neither outputfile nor a output format option are set,
            # default to printing .dot format to stdout. Kept for backward
            # compatibility.
            output = "dot"
        elif outputfile_ext == ".dot":
            output = "dot"
        elif outputfile_ext == ".json":
            output = "json"
        elif HAS_PYGRAPHVIZ:
            output = "pygraphviz"
        elif HAS_PYDOT:
            output = "pydot"
        else:
            raise CommandError(
                "Neither pygraphviz nor pydotplus could be found to generate the image."
                " To generate text output, use the --json or --dot options."
            )

        if options.get("rankdir") != "TB" and output not in [
            "pydot",
            "pygraphviz",
            "dot",
        ]:
            raise CommandError(
                "--rankdir is not supported for the chosen output format"
            )

        if options.get("ordering") and output not in ["pydot", "pygraphviz", "dot"]:
            raise CommandError(
                "--ordering is not supported for the chosen output format"
            )

        # Consistency check: Abort if --pygraphviz or --pydot options are set
        # but no outputfile is specified. Before 2.1.4 this silently fell back
        # to printind .dot format to stdout.
        if output in ["pydot", "pygraphviz"] and not outputfile:
            raise CommandError(
                "An output file (--output) must be specified when --pydot or "
                "--pygraphviz are set."
            )

        cli_options = " ".join(sys.argv[2:])
        graph_models = ModelGraph(args, cli_options=cli_options, **options)
        graph_models.generate_graph_data()

        if output == "json":
            graph_data = graph_models.get_graph_data(as_json=True)
            return self.render_output_json(graph_data, outputfile)

        graph_data = graph_models.get_graph_data(as_json=False)

        theme = options["theme"]
        template_name = os.path.join(
            "django_extensions", "graph_models", theme, "digraph.dot"
        )
        template = loader.get_template(template_name)

        graph_data = retheme(graph_data, app_style=options["app-style"])
        dotdata = generate_dot(graph_data, template=template)

        if output == "pygraphviz":
            return self.render_output_pygraphviz(dotdata, **options)
        if output == "pydot":
            return self.render_output_pydot(dotdata, **options)
        self.print_output(dotdata, outputfile)

    def print_output(self, dotdata, output_file=None):
        """Write model data to file or stdout in DOT (text) format."""
        if isinstance(dotdata, bytes):
            dotdata = dotdata.decode()

        if output_file:
            with open(output_file, "wt") as dot_output_f:
                dot_output_f.write(dotdata)
        else:
            self.stdout.write(dotdata)

    def render_output_json(self, graph_data, output_file=None):
        """Write model data to file or stdout in JSON format."""
        if output_file:
            with open(output_file, "wt") as json_output_f:
                json.dump(graph_data, json_output_f)
        else:
            self.stdout.write(json.dumps(graph_data))

    def render_output_pygraphviz(self, dotdata, **kwargs):
        """Render model data as image using pygraphviz."""
        if not HAS_PYGRAPHVIZ:
            raise CommandError("You need to install pygraphviz python module")

        version = pygraphviz.__version__.rstrip("-svn")
        try:
            if tuple(int(v) for v in version.split(".")) < (0, 36):
                # HACK around old/broken AGraph before version 0.36
                #   (ubuntu ships with this old version)
                tmpfile = tempfile.NamedTemporaryFile()
                tmpfile.write(dotdata)
                tmpfile.seek(0)
                dotdata = tmpfile.name
        except ValueError:
            pass

        graph = pygraphviz.AGraph(dotdata)
        graph.layout(prog=kwargs["layout"])
        graph.draw(kwargs["outputfile"])

    def render_output_pydot(self, dotdata, **kwargs):
        """Render model data as image using pydot."""
        if not HAS_PYDOT:
            raise CommandError("You need to install pydot python module")

        graph = pydot.graph_from_dot_data(dotdata)
        if not graph:
            raise CommandError("pydot returned an error")
        if isinstance(graph, (list, tuple)):
            if len(graph) > 1:
                sys.stderr.write(
                    "Found more then one graph, rendering only the first one.\n"
                )
            graph = graph[0]

        output_file = kwargs["outputfile"]
        formats = [
            "bmp",
            "canon",
            "cmap",
            "cmapx",
            "cmapx_np",
            "dot",
            "dia",
            "emf",
            "em",
            "fplus",
            "eps",
            "fig",
            "gd",
            "gd2",
            "gif",
            "gv",
            "imap",
            "imap_np",
            "ismap",
            "jpe",
            "jpeg",
            "jpg",
            "metafile",
            "pdf",
            "pic",
            "plain",
            "plain-ext",
            "png",
            "pov",
            "ps",
            "ps2",
            "svg",
            "svgz",
            "tif",
            "tiff",
            "tk",
            "vml",
            "vmlz",
            "vrml",
            "wbmp",
            "webp",
            "xdot",
        ]
        ext = output_file[output_file.rfind(".") + 1 :]
        format_ = ext if ext in formats else "raw"
        graph.write(output_file, format=format_)



# -*- coding: utf-8 -*-
import json
from operator import itemgetter
from pathlib import Path

from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.recorder import MigrationRecorder
from django.utils import timezone

from django_extensions.management.utils import signalcommand

DEFAULT_FILENAME = "managestate.json"
DEFAULT_STATE = "default"


class Command(BaseCommand):
    help = "Manage database state in the convenient way."
    _applied_migrations = None
    migrate_args: dict
    migrate_options: dict
    filename: str
    verbosity: int
    database: str
    conn: BaseDatabaseWrapper

    def add_arguments(self, parser):
        parser.add_argument(
            "action",
            choices=("dump", "load"),
            help="An action to do. "
            "Dump action saves applied migrations to a file. "
            "Load action applies migrations specified in a file.",
        )
        parser.add_argument(
            "state",
            nargs="?",
            default=DEFAULT_STATE,
            help="A name of a state. Usually a name of a git branch."
            f'Defaults to "{DEFAULT_STATE}"',
        )
        parser.add_argument(
            "-d",
            "--database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates a database to synchronize. "
            f'Defaults to the "{DEFAULT_DB_ALIAS}" database.',
        )
        parser.add_argument(
            "-f",
            "--filename",
            default=DEFAULT_FILENAME,
            help=f'A file to write to. Defaults to "{DEFAULT_FILENAME}"',
        )

        # migrate command arguments
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            help='The argument for "migrate" command. '
            "Tells Django to NOT prompt the user for input of any kind.",
        )
        parser.add_argument(
            "--fake",
            action="store_true",
            help='The argument for "migrate" command. '
            "Mark migrations as run without actually running them.",
        )
        parser.add_argument(
            "--fake-initial",
            action="store_true",
            help='The argument for "migrate" command. '
            "Detect if tables already exist and fake-apply initial migrations if so. "
            "Make sure that the current database schema matches your initial migration "
            "before using this flag. "
            "Django will only check for an existing table name.",
        )
        parser.add_argument(
            "--plan",
            action="store_true",
            help='The argument for "migrate" command. '
            "Shows a list of the migration actions that will be performed.",
        )
        parser.add_argument(
            "--run-syncdb",
            action="store_true",
            help='The argument for "migrate" command. '
            "Creates tables for apps without migrations.",
        )
        parser.add_argument(
            "--check",
            action="store_true",
            dest="check_unapplied",
            help='The argument for "migrate" command. '
            "Exits with a non-zero status if unapplied migrations exist.",
        )

    @signalcommand
    def handle(self, action, database, filename, state, *args, **options):
        self.migrate_args = args
        self.migrate_options = options
        self.verbosity = options["verbosity"]
        self.conn = connections[database]
        self.database = database
        self.filename = filename
        getattr(self, action)(state)

    def dump(self, state: str):
        """Save applied migrations to a file."""
        migrated_apps = self.get_migrated_apps()
        migrated_apps.update(self.get_applied_migrations())
        self.write({state: migrated_apps})
        self.stdout.write(
            self.style.SUCCESS(
                f'Migrations for state "{state}" have been successfully '
                f"saved to {self.filename}."
            )
        )

    def load(self, state: str):
        """Apply migrations from a file."""
        migrations = self.read().get(state)
        if migrations is None:
            raise CommandError(f"No such state saved: {state}")

        kwargs = {
            **self.migrate_options,
            "database": self.database,
            "verbosity": self.verbosity - 1 if self.verbosity > 1 else 0,
        }

        for app, migration in migrations.items():
            if self.is_applied(app, migration):
                continue

            if self.verbosity > 1:
                self.stdout.write(
                    self.style.WARNING(f'Applying migrations for "{app}"')
                )
            args = (app, migration, *self.migrate_args)
            call_command("migrate", *args, **kwargs)

        self.stdout.write(
            self.style.SUCCESS(
                f'Migrations for "{state}" have been successfully applied.'
            )
        )

    def get_migrated_apps(self) -> dict:
        """Installed apps having migrations."""
        apps = MigrationLoader(self.conn).migrated_apps
        migrated_apps = dict.fromkeys(apps, "zero")
        if self.verbosity > 1:
            self.stdout.write(
                "Apps having migrations: " + ", ".join(sorted(migrated_apps))
            )
        return migrated_apps

    def get_applied_migrations(self) -> dict:
        """Installed apps with last applied migrations."""
        if self._applied_migrations:
            return self._applied_migrations

        migrations = MigrationRecorder(self.conn).applied_migrations()
        last_applied = sorted(migrations.keys(), key=itemgetter(1))

        self._applied_migrations = dict(last_applied)
        return self._applied_migrations

    def is_applied(self, app: str, migration: str) -> bool:
        """Check whether a migration for an app is applied or not."""
        applied = self.get_applied_migrations().get(app)
        if applied == migration:
            if self.verbosity > 1:
                self.stdout.write(
                    self.style.WARNING(f'Migrations for "{app}" are already applied.')
                )
            return True
        return False

    def read(self) -> dict:
        """Get saved state from the file."""
        path = Path(self.filename)
        if not path.exists() or not path.is_file():
            raise CommandError(f"No such file: {self.filename}")

        with open(self.filename) as file:
            return json.load(file)

    def write(self, data: dict):
        """Write new data to the file using existent one."""
        try:
            saved = self.read()
        except CommandError:
            saved = {}

        saved.update(data, updated_at=str(timezone.now()))
        with open(self.filename, "w") as file:
            json.dump(saved, file, indent=2, sort_keys=True)



# -*- coding: utf-8 -*-
import os
import re

from django.conf import settings
from django.core.management.base import BaseCommand

from django_extensions.compat import get_template_setting
from django_extensions.management.utils import signalcommand

ANNOTATION_RE = re.compile(
    r"\{?#[\s]*?(TODO|FIXME|BUG|HACK|WARNING|NOTE|XXX)[\s:]?(.+)"
)
ANNOTATION_END_RE = re.compile(r"(.*)#\}(.*)")


class Command(BaseCommand):
    help = "Show all annotations like TODO, FIXME, BUG, HACK, WARNING, NOTE or XXX "
    "in your py and HTML files."
    label = "annotation tag (TODO, FIXME, BUG, HACK, WARNING, NOTE, XXX)"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--tag", dest="tag", help="Search for specific tags only", action="append"
        )

    @signalcommand
    def handle(self, *args, **options):
        # don't add django internal code
        apps = [
            app.replace(".", "/")
            for app in filter(
                lambda app: not app.startswith("django.contrib"),
                settings.INSTALLED_APPS,
            )
        ]
        template_dirs = get_template_setting("DIRS", [])
        base_dir = getattr(settings, "BASE_DIR")
        if template_dirs:
            apps += template_dirs
        for app_dir in apps:
            if base_dir:
                app_dir = os.path.join(base_dir, app_dir)
            for top, dirs, files in os.walk(app_dir):
                for fn in files:
                    if os.path.splitext(fn)[1] in (".py", ".html"):
                        fpath = os.path.join(top, fn)
                        annotation_lines = []
                        with open(fpath, "r") as fd:
                            i = 0
                            for line in fd.readlines():
                                i += 1
                                if ANNOTATION_RE.search(line):
                                    tag, msg = ANNOTATION_RE.findall(line)[0]
                                    if options["tag"]:
                                        if tag not in map(
                                            str.upper, map(str, options["tag"])
                                        ):
                                            break

                                    if ANNOTATION_END_RE.search(msg.strip()):
                                        msg = ANNOTATION_END_RE.findall(msg.strip())[0][
                                            0
                                        ]

                                    annotation_lines.append(
                                        "[%3s] %-5s %s" % (i, tag, msg.strip())
                                    )
                            if annotation_lines:
                                self.stdout.write("%s:" % fpath)
                                for annotation in annotation_lines:
                                    self.stdout.write("  * %s" % annotation)
                                self.stdout.write("")



# -*- coding: utf-8 -*-
"""
set_fake_emails.py

    Give all users a new email account. Useful for testing in a
    development environment. As such, this command is only available when
    setting.DEBUG is True.

"""

from typing import List

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError

from django_extensions.management.utils import signalcommand

DEFAULT_FAKE_EMAIL = "%(username)s@example.com"


class Command(BaseCommand):
    help = (
        "DEBUG only: give all users a new email based on their account data "
        '("%s" by default). '
        "Possible parameters are: username, first_name, last_name"
    ) % (DEFAULT_FAKE_EMAIL,)
    requires_system_checks: List[str] = []

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--email",
            dest="default_email",
            default=DEFAULT_FAKE_EMAIL,
            help="Use this as the new email format.",
        )
        parser.add_argument(
            "-a",
            "--no-admin",
            action="store_true",
            dest="no_admin",
            default=False,
            help="Do not change administrator accounts",
        )
        parser.add_argument(
            "-s",
            "--no-staff",
            action="store_true",
            dest="no_staff",
            default=False,
            help="Do not change staff accounts",
        )
        parser.add_argument(
            "--include",
            dest="include_regexp",
            default=None,
            help="Include usernames matching this regexp.",
        )
        parser.add_argument(
            "--exclude",
            dest="exclude_regexp",
            default=None,
            help="Exclude usernames matching this regexp.",
        )
        parser.add_argument(
            "--include-groups",
            dest="include_groups",
            default=None,
            help=(
                "Include users matching this group. "
                "(use comma separation for multiple groups)"
            ),
        )
        parser.add_argument(
            "--exclude-groups",
            dest="exclude_groups",
            default=None,
            help=(
                "Exclude users matching this group. "
                "(use comma separation for multiple groups)"
            ),
        )

    @signalcommand
    def handle(self, *args, **options):
        if not settings.DEBUG:
            raise CommandError("Only available in debug mode")

        from django.contrib.auth.models import Group

        email = options["default_email"]
        include_regexp = options["include_regexp"]
        exclude_regexp = options["exclude_regexp"]
        include_groups = options["include_groups"]
        exclude_groups = options["exclude_groups"]
        no_admin = options["no_admin"]
        no_staff = options["no_staff"]

        User = get_user_model()
        users = User.objects.all()
        if no_admin:
            users = users.exclude(is_superuser=True)
        if no_staff:
            users = users.exclude(is_staff=True)
        if exclude_groups:
            groups = Group.objects.filter(name__in=exclude_groups.split(","))
            if groups:
                users = users.exclude(groups__in=groups)
            else:
                raise CommandError("No groups matches filter: %s" % exclude_groups)
        if include_groups:
            groups = Group.objects.filter(name__in=include_groups.split(","))
            if groups:
                users = users.filter(groups__in=groups)
            else:
                raise CommandError("No groups matches filter: %s" % include_groups)
        if exclude_regexp:
            users = users.exclude(username__regex=exclude_regexp)
        if include_regexp:
            users = users.filter(username__regex=include_regexp)
        for user in users:
            user.email = email % {
                "username": user.username,
                "first_name": user.first_name,
                "last_name": user.last_name,
            }
            user.save()
        print("Changed %d emails" % users.count())



# -*- coding: utf-8 -*-
import ast
import traceback
import warnings
import importlib

from typing import (  # NOQA
    Dict,
    List,
    Tuple,
    Union,
)

from django.apps.config import MODELS_MODULE_NAME
from django.utils.module_loading import import_string
from django_extensions.collision_resolvers import CollisionResolvingRunner
from django_extensions.import_subclasses import SubclassesFinder
from django_extensions.utils.deprecation import RemovedInNextVersionWarning


SHELL_PLUS_DJANGO_IMPORTS = [
    "from django.core.cache import cache",
    "from django.conf import settings",
    "from django.contrib.auth import get_user_model",
    "from django.db import transaction",
    "from django.db.models import Avg, Case, Count, F, Max, Min, Prefetch, Q, Sum, When",  # noqa: E501
    "from django.utils import timezone",
    "from django.urls import reverse",
    "from django.db.models import Exists, OuterRef, Subquery",
]


class ObjectImportError(Exception):
    pass


def get_app_name(mod_name):
    """
    Retrieve application name from models.py module path

    >>> get_app_name('testapp.models.foo')
    'testapp'

    'testapp' instead of 'some.testapp' for compatibility:
    >>> get_app_name('some.testapp.models.foo')
    'testapp'
    >>> get_app_name('some.models.testapp.models.foo')
    'testapp'
    >>> get_app_name('testapp.foo')
    'testapp'
    >>> get_app_name('some.testapp.foo')
    'testapp'
    """
    rparts = list(reversed(mod_name.split(".")))
    try:
        try:
            return rparts[rparts.index(MODELS_MODULE_NAME) + 1]
        except ValueError:
            # MODELS_MODULE_NAME ('models' string) is not found
            return rparts[1]
    except IndexError:
        # Some weird model naming scheme like in Sentry.
        return mod_name


def import_items(import_directives, style, quiet_load=False):
    """
    Import the items in import_directives and return a list of the imported items

    Each item in import_directives should be one of the following forms
        * a tuple like ('module.submodule', ('classname1', 'classname2')), which indicates a 'from module.submodule import classname1, classname2'
        * a tuple like ('module.submodule', 'classname1'), which indicates a 'from module.submodule import classname1'
        * a tuple like ('module.submodule', '*'), which indicates a 'from module.submodule import *'
        * a simple 'module.submodule' which indicates 'import module.submodule'.

    Returns a dict mapping the names to the imported items
    """  # noqa: E501
    imported_objects = {}

    for directive in import_directives:
        if isinstance(directive, str):
            directive = directive.strip()
        try:
            if isinstance(directive, str) and directive.startswith(
                ("from ", "import ")
            ):
                try:
                    node = ast.parse(directive)
                except Exception as exc:
                    if not quiet_load:
                        print(style.ERROR("Error parsing: %r %s" % (directive, exc)))
                        continue
                if not all(
                    isinstance(body, (ast.Import, ast.ImportFrom)) for body in node.body
                ):
                    if not quiet_load:
                        print(
                            style.ERROR(
                                "Only specify import statements: %r" % directive
                            )
                        )
                    continue

                if not quiet_load:
                    print(style.SQL_COLTYPE("%s" % directive))

                for body in node.body:
                    if isinstance(body, ast.Import):
                        for name in body.names:
                            asname = name.asname or name.name
                            imported_objects[asname] = importlib.import_module(
                                name.name
                            )
                    if isinstance(body, ast.ImportFrom):
                        imported_object = importlib.__import__(
                            body.module, {}, {}, [name.name for name in body.names]
                        )
                        for name in body.names:
                            asname = name.asname or name.name
                            try:
                                if name.name == "*":
                                    for k in dir(imported_object):
                                        imported_objects[k] = getattr(
                                            imported_object, k
                                        )
                                else:
                                    imported_objects[asname] = getattr(
                                        imported_object, name.name
                                    )
                            except AttributeError as exc:
                                print(
                                    (
                                        f"Couldn't find {name.name} in {body.module}. "
                                        f"Only found: {dir(imported_object)}"
                                    )
                                )
                                # raise
                                raise ImportError(exc)
            else:
                warnings.warn(
                    "Old style import definitions are deprecated. You should use the new style which is similar to normal Python imports. ",  # noqa: E501
                    RemovedInNextVersionWarning,
                    stacklevel=2,
                )

                if isinstance(directive, str):
                    imported_object = __import__(directive)
                    imported_objects[directive.split(".")[0]] = imported_object
                    if not quiet_load:
                        print(style.SQL_COLTYPE("import %s" % directive))
                    continue
                elif isinstance(directive, (list, tuple)) and len(directive) == 2:
                    if not isinstance(directive[0], str):
                        if not quiet_load:
                            print(
                                style.ERROR(
                                    (
                                        "Unable to import %r: module name must be of "
                                        "type string"
                                    )
                                    % directive[0]
                                )
                            )
                        continue

                    if isinstance(directive[1], (list, tuple)) and all(
                        isinstance(e, str) for e in directive[1]
                    ):
                        # Try the form of:
                        #   ('module.submodule', ('classname1', 'classname2'))
                        imported_object = __import__(directive[0], {}, {}, directive[1])
                        imported_names = []
                        for name in directive[1]:
                            try:
                                imported_objects[name] = getattr(imported_object, name)
                            except AttributeError:
                                if not quiet_load:
                                    print(
                                        style.ERROR(
                                            "Unable to import %r from %r: %r does not "
                                            "exist" % (name, directive[0], name)
                                        )
                                    )
                            else:
                                imported_names.append(name)
                        if not quiet_load:
                            print(
                                style.SQL_COLTYPE(
                                    "from %s import %s"
                                    % (directive[0], ", ".join(imported_names))
                                )
                            )
                    elif isinstance(directive[1], str):
                        # If it is a tuple, but the second item isn't a list, so we have
                        # something like ('module.submodule', 'classname1')
                        # Check for the special '*' to import all
                        if directive[1] == "*":
                            imported_object = __import__(
                                directive[0], {}, {}, directive[1]
                            )
                            for k in dir(imported_object):
                                imported_objects[k] = getattr(imported_object, k)
                            if not quiet_load:
                                print(
                                    style.SQL_COLTYPE("from %s import *" % directive[0])
                                )
                        else:
                            imported_object = getattr(
                                __import__(directive[0], {}, {}, [directive[1]]),
                                directive[1],
                            )
                            imported_objects[directive[1]] = imported_object
                            if not quiet_load:
                                print(
                                    style.SQL_COLTYPE(
                                        "from %s import %s"
                                        % (directive[0], directive[1])
                                    )
                                )
                    else:
                        if not quiet_load:
                            print(
                                style.ERROR(
                                    "Unable to import %r from %r: names must be strings"
                                    % (directive[1], directive[0])
                                )
                            )
                else:
                    if not quiet_load:
                        print(
                            style.ERROR(
                                "Unable to import %r: names must be strings" % directive
                            )
                        )
        except ImportError:
            if not quiet_load:
                print(style.ERROR("Unable to import %r" % directive))

    return imported_objects


def import_objects(options, style):
    from django.apps import apps
    from django import setup

    if not apps.ready:
        setup()

    from django.conf import settings

    dont_load_cli = options.get("dont_load", [])
    dont_load_conf = getattr(settings, "SHELL_PLUS_DONT_LOAD", [])
    dont_load = dont_load_cli + dont_load_conf
    dont_load_any_models = "*" in dont_load
    quiet_load = options.get("quiet_load")
    model_aliases = getattr(settings, "SHELL_PLUS_MODEL_ALIASES", {})
    app_prefixes = getattr(settings, "SHELL_PLUS_APP_PREFIXES", {})
    SHELL_PLUS_PRE_IMPORTS = getattr(settings, "SHELL_PLUS_PRE_IMPORTS", {})

    imported_objects = {}
    load_models = {}

    def get_dict_from_names_to_possible_models():  # type: () -> Dict[str, List[str]]
        """
        Collect dictionary from names to possible models. Model is represented as his full path.
        Name of model can be alias if SHELL_PLUS_MODEL_ALIASES or SHELL_PLUS_APP_PREFIXES is specified for this model.
        This dictionary is used by collision resolver.
        At this phase we can't import any models, because collision resolver can change results.
        :return: Dict[str, List[str]]. Key is name, value is list of full model's path's.
        """  # noqa: E501
        models_to_import = {}  # type: Dict[str, List[str]]
        for app_mod, models in sorted(load_models.items()):
            app_name = get_app_name(app_mod)
            app_aliases = model_aliases.get(app_name, {})
            prefix = app_prefixes.get(app_name)

            for model_name in sorted(models):
                if "%s.%s" % (app_name, model_name) in dont_load:
                    continue

                alias = app_aliases.get(model_name)

                if not alias:
                    if prefix:
                        alias = "%s_%s" % (prefix, model_name)
                    else:
                        alias = model_name

                models_to_import.setdefault(alias, [])
                models_to_import[alias].append("%s.%s" % (app_mod, model_name))
        return models_to_import

    def import_subclasses():
        base_classes_to_import = getattr(settings, "SHELL_PLUS_SUBCLASSES_IMPORT", [])  # type: List[Union[str, type]]
        if base_classes_to_import:
            if not quiet_load:
                print(style.SQL_TABLE("# Shell Plus Subclasses Imports"))
            perform_automatic_imports(
                SubclassesFinder(base_classes_to_import).collect_subclasses()
            )

    def import_models():
        """
        Perform collision resolving and imports all models.
        When collisions are resolved we can perform imports and print information's, because it is last phase.
        This function updates imported_objects dictionary.
        """  # noqa: E501
        modules_to_models = CollisionResolvingRunner().run_collision_resolver(
            get_dict_from_names_to_possible_models()
        )
        perform_automatic_imports(modules_to_models)

    def perform_automatic_imports(modules_to_classes):  # type: (Dict[str, List[Tuple[str, str]]]) -> ()
        """
        Import elements from given dictionary.
        :param modules_to_classes: dictionary from module name to tuple.
        First element of tuple is model name, second is model alias.
        If both elements are equal than element is imported without alias.
        """
        for full_module_path, models in modules_to_classes.items():
            model_labels = []
            for model_name, alias in sorted(models):
                try:
                    imported_objects[alias] = import_string(
                        "%s.%s" % (full_module_path, model_name)
                    )
                    if model_name == alias:
                        model_labels.append(model_name)
                    else:
                        model_labels.append("%s (as %s)" % (model_name, alias))
                except ImportError as e:
                    if options.get("traceback"):
                        traceback.print_exc()
                    if not options.get("quiet_load"):
                        print(
                            style.ERROR(
                                "Failed to import '%s' from '%s' reason: %s"
                                % (model_name, full_module_path, str(e))
                            )
                        )
            if not options.get("quiet_load"):
                print(
                    style.SQL_COLTYPE(
                        "from %s import %s"
                        % (full_module_path, ", ".join(model_labels))
                    )
                )

    def get_apps_and_models():
        for app in apps.get_app_configs():
            if app.models_module:
                yield app.models_module, app.get_models()

    mongoengine = False
    try:
        from mongoengine.base import _document_registry

        mongoengine = True
    except ImportError:
        pass

    # Perform pre-imports before any other imports
    if SHELL_PLUS_PRE_IMPORTS:
        if not quiet_load:
            print(style.SQL_TABLE("# Shell Plus User Pre Imports"))
        imports = import_items(SHELL_PLUS_PRE_IMPORTS, style, quiet_load=quiet_load)
        for k, v in imports.items():
            imported_objects[k] = v

    if mongoengine and not dont_load_any_models:
        for name, mod in _document_registry.items():
            name = name.split(".")[-1]
            app_name = get_app_name(mod.__module__)
            if app_name in dont_load or ("%s.%s" % (app_name, name)) in dont_load:
                continue

            load_models.setdefault(mod.__module__, [])
            load_models[mod.__module__].append(name)

    if not dont_load_any_models:
        for app_mod, app_models in get_apps_and_models():
            if not app_models:
                continue

            app_name = get_app_name(app_mod.__name__)
            if app_name in dont_load:
                continue

            for mod in app_models:
                if "%s.%s" % (app_name, mod.__name__) in dont_load:
                    continue

                if mod.__module__:
                    # Only add the module to the dict if `__module__` is not empty.
                    load_models.setdefault(mod.__module__, [])
                    load_models[mod.__module__].append(mod.__name__)

    import_subclasses()
    if not quiet_load:
        print(
            style.SQL_TABLE("# Shell Plus Model Imports%s")
            % (" SKIPPED" if dont_load_any_models else "")
        )

    import_models()

    # Imports often used from Django
    if getattr(settings, "SHELL_PLUS_DJANGO_IMPORTS", True):
        if not quiet_load:
            print(style.SQL_TABLE("# Shell Plus Django Imports"))
        imports = import_items(SHELL_PLUS_DJANGO_IMPORTS, style, quiet_load=quiet_load)
        for k, v in imports.items():
            imported_objects[k] = v

    SHELL_PLUS_IMPORTS = getattr(settings, "SHELL_PLUS_IMPORTS", {})
    if SHELL_PLUS_IMPORTS:
        if not quiet_load:
            print(style.SQL_TABLE("# Shell Plus User Imports"))
        imports = import_items(SHELL_PLUS_IMPORTS, style, quiet_load=quiet_load)
        for k, v in imports.items():
            imported_objects[k] = v

    # Perform post-imports after any other imports
    SHELL_PLUS_POST_IMPORTS = getattr(settings, "SHELL_PLUS_POST_IMPORTS", {})
    if SHELL_PLUS_POST_IMPORTS:
        if not quiet_load:
            print(style.SQL_TABLE("# Shell Plus User Post Imports"))
        imports = import_items(SHELL_PLUS_POST_IMPORTS, style, quiet_load=quiet_load)
        for k, v in imports.items():
            imported_objects[k] = v

    return imported_objects



# -*- coding: utf-8 -*-
import sys

from django.core.management.base import LabelCommand
from django.template import TemplateDoesNotExist, loader

from django_extensions.management.utils import signalcommand


class Command(LabelCommand):
    help = "Finds the location of the given template by resolving its path"
    args = "[template_path]"
    label = "template path"

    @signalcommand
    def handle_label(self, template_path, **options):
        try:
            template = loader.get_template(template_path).template
        except TemplateDoesNotExist:
            sys.stderr.write("No template found\n")
        else:
            sys.stdout.write(self.style.SUCCESS((template.name)))



# -*- coding: utf-8 -*-
import socket

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.apps import apps

from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    help = "Set parameters of the default django.contrib.sites Site"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--name", dest="site_name", default=None, help="Use this as site name."
        )
        parser.add_argument(
            "--domain",
            dest="site_domain",
            default=None,
            help="Use this as site domain.",
        )
        parser.add_argument(
            "--system-fqdn",
            dest="set_as_system_fqdn",
            default=False,
            action="store_true",
            help="Use the systems FQDN (Fully Qualified Domain Name) as name "
            "and domain. Can be used in combination with --name",
        )

    @signalcommand
    def handle(self, *args, **options):
        if not apps.is_installed("django.contrib.sites"):
            raise CommandError("The sites framework is not installed.")

        from django.contrib.sites.models import Site

        try:
            site = Site.objects.get(pk=settings.SITE_ID)
        except Site.DoesNotExist:
            raise CommandError(
                "Default site with pk=%s does not exist" % settings.SITE_ID
            )
        else:
            name = options["site_name"]
            domain = options["site_domain"]
            set_as_system_fqdn = options["set_as_system_fqdn"]
            if all([domain, set_as_system_fqdn]):
                raise CommandError(
                    "The set_as_system_fqdn cannot be used with domain option."
                )  # noqa
            if set_as_system_fqdn:
                domain = socket.getfqdn()
                if not domain:
                    raise CommandError("Cannot find systems FQDN")
                if name is None:
                    name = domain

            update_kwargs = {}
            if name and name != site.name:
                update_kwargs["name"] = name

            if domain and domain != site.domain:
                update_kwargs["domain"] = domain

            if update_kwargs:
                Site.objects.filter(pk=settings.SITE_ID).update(**update_kwargs)
                site = Site.objects.get(pk=settings.SITE_ID)
                print(
                    "Updated default site. You might need to restart django as sites"
                    " are cached aggressively."
                )
            else:
                print("Nothing to update (need --name, --domain and/or --system-fqdn)")

            print("Default Site:")
            print("\tid = %s" % site.id)
            print("\tname = %s" % site.name)
            print("\tdomain = %s" % site.domain)



# -*- coding: utf-8 -*-
import os
import sys
from typing import List

from django.core.management.base import AppCommand

from django_extensions.management.utils import _make_writeable, signalcommand


class Command(AppCommand):
    help = "Creates a Django template tags directory structure for the given app name "
    "in the apps's directory"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--name",
            "-n",
            action="store",
            dest="tag_library_name",
            default="appname_tags",
            help="The name to use for the template tag base name. "
            "Defaults to `appname`_tags.",
        )

    requires_system_checks: List[str] = []
    # Can't import settings during this command, because they haven't
    # necessarily been created.
    can_import_settings = True

    @signalcommand
    def handle_app_config(self, app_config, **options):
        app_dir = app_config.path
        tag_library_name = options["tag_library_name"]
        if tag_library_name == "appname_tags":
            tag_library_name = "%s_tags" % os.path.basename(app_dir)
        copy_template("template_tags_template", app_dir, tag_library_name)


def copy_template(template_name, copy_to, tag_library_name):
    """Copy the specified template directory to the copy_to location"""
    import django_extensions
    import shutil

    template_dir = os.path.join(django_extensions.__path__[0], "conf", template_name)

    # walk the template structure and copies it
    for d, subdirs, files in os.walk(template_dir):
        relative_dir = d[len(template_dir) + 1 :]
        if relative_dir and not os.path.exists(os.path.join(copy_to, relative_dir)):
            os.mkdir(os.path.join(copy_to, relative_dir))
        for i, subdir in enumerate(subdirs):
            if subdir.startswith("."):
                del subdirs[i]
        for f in files:
            if f.endswith(".pyc") or f.startswith(".DS_Store"):
                continue
            path_old = os.path.join(d, f)
            path_new = os.path.join(
                copy_to, relative_dir, f.replace("sample", tag_library_name)
            )
            if os.path.exists(path_new):
                path_new = os.path.join(copy_to, relative_dir, f)
                if os.path.exists(path_new):
                    continue
            path_new = path_new.rstrip(".tmpl")
            fp_old = open(path_old, "r")
            fp_new = open(path_new, "w")
            fp_new.write(fp_old.read())
            fp_old.close()
            fp_new.close()
            try:
                shutil.copymode(path_old, path_new)
                _make_writeable(path_new)
            except OSError:
                sys.stderr.write(
                    "Notice: Couldn't set permission bits on %s. You're probably using an uncommon filesystem setup. No problem.\n"  # noqa: E501
                    % path_new
                )



# -*- coding: utf-8 -*-
import inspect
import sys
from abc import abstractmethod, ABCMeta
from typing import (  # NOQA
    Dict,
    List,
    Optional,
    Tuple,
)

from django.utils.module_loading import import_string


class BaseCR(metaclass=ABCMeta):
    """
    Abstract base collision resolver. All collision resolvers needs to inherit from this class.
    To write custom collision resolver you need to overwrite resolve_collisions function.
    It receives Dict[str, List[str]], where key is model name and values are full model names
    (full model name means: module + model_name).
    You should return Dict[str, str], where key is model name and value is full model name.
    """  # noqa: E501

    @classmethod
    def get_app_name_and_model(cls, full_model_path):  # type: (str) -> Tuple[str, str]
        model_class = import_string(full_model_path)
        return model_class._meta.app_config.name, model_class.__name__

    @abstractmethod
    def resolve_collisions(self, namespace):  # type: (Dict[str, List[str]]) -> Dict[str, str]
        pass


class LegacyCR(BaseCR):
    """
    Default collision resolver.

    Model from last application in alphabetical order is selected.
    """

    def resolve_collisions(self, namespace):
        result = {}
        for name, models in namespace.items():
            result[name] = models[-1]
        return result


class AppsOrderCR(LegacyCR, metaclass=ABCMeta):
    APP_PRIORITIES = None  # type: List[str]

    def resolve_collisions(self, namespace):
        assert self.APP_PRIORITIES is not None, (
            "You must define APP_PRIORITIES in your resolver class!"
        )
        result = {}
        for name, models in namespace.items():
            if len(models) > 0:
                sorted_models = self._sort_models_depending_on_priorities(models)
                result[name] = sorted_models[0][1]
        return result

    def _sort_models_depending_on_priorities(self, models):  # type: (List[str]) -> List[Tuple[int, str]]
        models_with_priorities = []
        for model in models:
            try:
                app_name, _ = self.get_app_name_and_model(model)
                position = self.APP_PRIORITIES.index(app_name)
            except (ImportError, ValueError):
                position = sys.maxsize
            models_with_priorities.append((position, model))
        return sorted(models_with_priorities)


class InstalledAppsOrderCR(AppsOrderCR):
    """
    Collision resolver which selects first model from INSTALLED_APPS.
    You can set your own app priorities list by subclassing him and overwriting APP_PRIORITIES field.
    This collision resolver will select model from first app on this list.
    If both app's are absent on this list, resolver will choose model from first app in alphabetical order.
    """  # noqa: E501

    @property
    def APP_PRIORITIES(self):
        from django.conf import settings

        return getattr(settings, "INSTALLED_APPS", [])


class PathBasedCR(LegacyCR, metaclass=ABCMeta):
    """
    Abstract resolver which transforms full model name into alias.
    To use him you need to overwrite transform_import function
    which should have one parameter. It will be full model name.
    It should return valid alias as str instance.
    """

    @abstractmethod
    def transform_import(self, module_path):  # type: (str) -> str
        pass

    def resolve_collisions(self, namespace):
        base_imports = super(PathBasedCR, self).resolve_collisions(namespace)
        for name, models in namespace.items():
            if len(models) <= 1:
                continue
            for model in models:
                new_name = self.transform_import(model)
                assert isinstance(new_name, str), (
                    "result of transform_import must be str!"
                )
                base_imports[new_name] = model
        return base_imports


class FullPathCR(PathBasedCR):
    """
    Collision resolver which transform full model name to alias by changing dots to underscores.
    He also removes 'models' part of alias, because all models are in models.py files.
    Model from last application in alphabetical order is selected.
    """  # noqa: E501

    def transform_import(self, module_path):
        module, model = module_path.rsplit(".models", 1)
        module_path = module + model
        return module_path.replace(".", "_")


class AppNameCR(PathBasedCR, metaclass=ABCMeta):
    """
    Abstract collision resolver which transform pair (app name, model_name) to alias by changing dots to underscores.
    You must define MODIFICATION_STRING which should be string to format with two keyword arguments:
    app_name and model_name. For example: "{app_name}_{model_name}".
    Model from last application in alphabetical order is selected.
    """  # noqa: E501

    MODIFICATION_STRING = None  # type: Optional[str]

    def transform_import(self, module_path):
        assert self.MODIFICATION_STRING is not None, (
            "You must define MODIFICATION_STRING in your resolver class!"
        )
        app_name, model_name = self.get_app_name_and_model(module_path)
        app_name = app_name.replace(".", "_")
        return self.MODIFICATION_STRING.format(app_name=app_name, model_name=model_name)


class AppNamePrefixCR(AppNameCR):
    """
    Collision resolver which transform pair (app name, model_name) to alias "{app_name}_{model_name}".
    Model from last application in alphabetical order is selected.
    Result is different than FullPathCR, when model has app_label other than current app.
    """  # noqa: E501

    MODIFICATION_STRING = "{app_name}_{model_name}"


class AppNameSuffixCR(AppNameCR):
    """
    Collision resolver which transform pair (app name, model_name) to alias "{model_name}_{app_name}"
    Model from last application in alphabetical order is selected.
    """  # noqa: E501

    MODIFICATION_STRING = "{model_name}_{app_name}"


class AppNamePrefixCustomOrderCR(AppNamePrefixCR, InstalledAppsOrderCR):
    """
    Collision resolver which is mixin of AppNamePrefixCR and InstalledAppsOrderCR.
    In case of collisions he sets aliases like AppNamePrefixCR, but sets default model using InstalledAppsOrderCR.
    """  # noqa: E501

    pass


class AppNameSuffixCustomOrderCR(AppNameSuffixCR, InstalledAppsOrderCR):
    """
    Collision resolver which is mixin of AppNameSuffixCR and InstalledAppsOrderCR.
    In case of collisions he sets aliases like AppNameSuffixCR, but sets default model using InstalledAppsOrderCR.
    """  # noqa: E501

    pass


class FullPathCustomOrderCR(FullPathCR, InstalledAppsOrderCR):
    """
    Collision resolver which is mixin of FullPathCR and InstalledAppsOrderCR.
    In case of collisions he sets aliases like FullPathCR, but sets default model using InstalledAppsOrderCR.
    """  # noqa: E501

    pass


class AppLabelCR(PathBasedCR, metaclass=ABCMeta):
    """
    Abstract collision resolver which transform pair (app_label, model_name) to alias.
    You must define MODIFICATION_STRING which should be string to format with two keyword arguments:
    app_label and model_name. For example: "{app_label}_{model_name}".
    This is different from AppNameCR when the app is nested with several level of namespace:
    Gives sites_Site instead of django_contrib_sites_Site
    Model from last application in alphabetical order is selected.
    """  # noqa: E501

    MODIFICATION_STRING = None  # type: Optional[str]

    def transform_import(self, module_path):
        assert self.MODIFICATION_STRING is not None, (
            "You must define MODIFICATION_STRING in your resolver class!"
        )
        model_class = import_string(module_path)
        app_label, model_name = model_class._meta.app_label, model_class.__name__
        return self.MODIFICATION_STRING.format(
            app_label=app_label, model_name=model_name
        )


class AppLabelPrefixCR(AppLabelCR):
    """
    Collision resolver which transform pair (app_label, model_name) to alias "{app_label}_{model_name}".
    Model from last application in alphabetical order is selected.
    """  # noqa: E501

    MODIFICATION_STRING = "{app_label}_{model_name}"


class AppLabelSuffixCR(AppLabelCR):
    """
    Collision resolver which transform pair (app_label, model_name) to alias "{model_name}_{app_label}".
    Model from last application in alphabetical order is selected.
    """  # noqa: E501

    MODIFICATION_STRING = "{model_name}_{app_label}"


class CollisionResolvingRunner:
    def __init__(self):
        pass

    def run_collision_resolver(self, models_to_import):
        # type: (Dict[str, List[str]]) -> Dict[str, List[Tuple[str, str]]]
        dictionary_of_names = self._get_dictionary_of_names(models_to_import)  # type: Dict[str, str]
        return self._get_dictionary_of_modules(dictionary_of_names)

    @classmethod
    def _get_dictionary_of_names(cls, models_to_import):  # type: (Dict[str, List[str]]) -> (Dict[str, str])
        from django.conf import settings

        collision_resolver_class = import_string(
            getattr(
                settings,
                "SHELL_PLUS_MODEL_IMPORTS_RESOLVER",
                "django_extensions.collision_resolvers.LegacyCR",
            )
        )

        cls._assert_is_collision_resolver_class_correct(collision_resolver_class)
        result = collision_resolver_class().resolve_collisions(models_to_import)
        cls._assert_is_collision_resolver_result_correct(result)

        return result

    @classmethod
    def _assert_is_collision_resolver_result_correct(cls, result):
        assert isinstance(result, dict), (
            "Result of resolve_collisions function must be a dict!"
        )
        for key, value in result.items():
            assert isinstance(key, str), (
                "key in collision resolver result should be str not %s" % key
            )
            assert isinstance(value, str), (
                "value in collision resolver result should be str not %s" % value
            )

    @classmethod
    def _assert_is_collision_resolver_class_correct(cls, collision_resolver_class):
        assert inspect.isclass(collision_resolver_class) and issubclass(
            collision_resolver_class, BaseCR
        ), "SHELL_PLUS_MODEL_IMPORTS_RESOLVER must be subclass of BaseCR!"
        assert (
            len(
                inspect.getfullargspec(collision_resolver_class.resolve_collisions).args
            )
            == 2
        ), "resolve_collisions function must take one argument!"

    @classmethod
    def _get_dictionary_of_modules(cls, dictionary_of_names):
        # type: (Dict[str, str]) -> Dict[str, List[Tuple[str, str]]]
        dictionary_of_modules = {}  # type: Dict[str, List[Tuple[str, str]]]
        for alias, model in dictionary_of_names.items():
            module_path, model_name = model.rsplit(".", 1)
            dictionary_of_modules.setdefault(module_path, [])
            dictionary_of_modules[module_path].append((model_name, alias))
        return dictionary_of_modules



# -*- coding: utf-8 -*-
import logging

from django.apps import apps
from django.core.management.base import BaseCommand

from django_extensions.management.jobs import get_jobs, print_jobs
from django_extensions.management.utils import setup_logger, signalcommand

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Runs scheduled maintenance jobs."

    when_options = [
        "minutely",
        "quarter_hourly",
        "hourly",
        "daily",
        "weekly",
        "monthly",
        "yearly",
    ]

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "when", nargs="?", help="options: %s" % ", ".join(self.when_options)
        )
        parser.add_argument(
            "--list",
            "-l",
            action="store_true",
            dest="list_jobs",
            default=False,
            help="List all jobs with their description",
        )

    def usage_msg(self):
        print("%s Please specify: %s" % (self.help, ", ".join(self.when_options)))

    def runjobs(self, when, options):
        verbosity = options["verbosity"]
        jobs = get_jobs(when, only_scheduled=True)
        for app_name, job_name in sorted(jobs.keys()):
            job = jobs[(app_name, job_name)]
            if verbosity > 1:
                logger.info("Executing %s job: %s (app: %s)", when, job_name, app_name)
            try:
                job().execute()
            except Exception:
                logger.exception(
                    "ERROR OCCURED IN JOB: %s (APP: %s)", job_name, app_name
                )

    def runjobs_by_signals(self, when, options):
        """Run jobs from the signals"""
        # Thanks for Ian Holsman for the idea and code
        from django_extensions.management import signals
        from django.conf import settings

        verbosity = options["verbosity"]
        for app_name in settings.INSTALLED_APPS:
            try:
                __import__(app_name + ".management", "", "", [""])
            except ImportError:
                pass

        for app in (
            app.models_module for app in apps.get_app_configs() if app.models_module
        ):
            if verbosity > 1:
                app_name = ".".join(app.__name__.rsplit(".")[:-1])
                print("Sending %s job signal for: %s" % (when, app_name))
            if when == "minutely":
                signals.run_minutely_jobs.send(sender=app, app=app)
            elif when == "quarter_hourly":
                signals.run_quarter_hourly_jobs.send(sender=app, app=app)
            elif when == "hourly":
                signals.run_hourly_jobs.send(sender=app, app=app)
            elif when == "daily":
                signals.run_daily_jobs.send(sender=app, app=app)
            elif when == "weekly":
                signals.run_weekly_jobs.send(sender=app, app=app)
            elif when == "monthly":
                signals.run_monthly_jobs.send(sender=app, app=app)
            elif when == "yearly":
                signals.run_yearly_jobs.send(sender=app, app=app)

    @signalcommand
    def handle(self, *args, **options):
        when = options["when"]

        setup_logger(logger, self.stdout)

        if options["list_jobs"]:
            print_jobs(when, only_scheduled=True, show_when=True, show_appname=True)
        elif when in self.when_options:
            self.runjobs(when, options)
            self.runjobs_by_signals(when, options)
        else:
            self.usage_msg()



# -*- coding: utf-8 -*-
import time
import traceback
from contextlib import contextmanager

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.backends import utils

from django_extensions.settings import DEFAULT_PRINT_SQL_TRUNCATE_CHARS


@contextmanager
def monkey_patch_cursordebugwrapper(
    print_sql=None,
    print_sql_location=False,
    truncate=None,
    logger=print,
    confprefix="DJANGO_EXTENSIONS",
):
    if not print_sql:
        yield
    else:
        if truncate is None:
            truncate = getattr(
                settings,
                "%s_PRINT_SQL_TRUNCATE" % confprefix,
                DEFAULT_PRINT_SQL_TRUNCATE_CHARS,
            )

        sqlparse = None
        if getattr(settings, "%s_SQLPARSE_ENABLED" % confprefix, True):
            try:
                import sqlparse

                sqlparse_format_kwargs_defaults = dict(
                    reindent_aligned=True,
                    truncate_strings=500,
                )
                sqlparse_format_kwargs = getattr(
                    settings,
                    "%s_SQLPARSE_FORMAT_KWARGS" % confprefix,
                    sqlparse_format_kwargs_defaults,
                )
            except ImportError:
                sqlparse = None

        pygments = None
        if getattr(settings, "%s_PYGMENTS_ENABLED" % confprefix, True):
            try:
                import pygments.lexers
                import pygments.formatters

                pygments_formatter = getattr(
                    settings,
                    "%s_PYGMENTS_FORMATTER" % confprefix,
                    pygments.formatters.TerminalFormatter,
                )
                pygments_formatter_kwargs = getattr(
                    settings, "%s_PYGMENTS_FORMATTER_KWARGS" % confprefix, {}
                )
            except ImportError:
                pass

        class PrintQueryWrapperMixin:
            def execute(self, sql, params=()):
                starttime = time.time()
                try:
                    return utils.CursorWrapper.execute(self, sql, params)
                finally:
                    execution_time = time.time() - starttime
                    raw_sql = self.db.ops.last_executed_query(self.cursor, sql, params)
                    if truncate:
                        raw_sql = raw_sql[:truncate]

                    if sqlparse:
                        raw_sql = sqlparse.format(raw_sql, **sqlparse_format_kwargs)

                    if pygments:
                        raw_sql = pygments.highlight(
                            raw_sql,
                            pygments.lexers.get_lexer_by_name("sql"),
                            pygments_formatter(**pygments_formatter_kwargs),
                        )

                    logger(raw_sql)
                    logger(
                        "Execution time: %.6fs [Database: %s]"
                        % (execution_time, self.db.alias)
                    )
                    if print_sql_location:
                        logger("Location of SQL Call:")
                        logger("".join(traceback.format_stack()))

        _CursorDebugWrapper = utils.CursorDebugWrapper

        class PrintCursorQueryWrapper(PrintQueryWrapperMixin, _CursorDebugWrapper):
            pass

        try:
            from django.db import connections

            _force_debug_cursor = {}
            for connection_name in connections:
                _force_debug_cursor[connection_name] = connections[
                    connection_name
                ].force_debug_cursor
        except Exception:
            connections = None

        utils.CursorDebugWrapper = PrintCursorQueryWrapper

        postgresql_base = None
        try:
            from django.db.backends.postgresql import base as postgresql_base

            _PostgreSQLCursorDebugWrapper = postgresql_base.CursorDebugWrapper

            class PostgreSQLPrintCursorDebugWrapper(
                PrintQueryWrapperMixin, _PostgreSQLCursorDebugWrapper
            ):
                pass
        except (ImproperlyConfigured, TypeError):
            postgresql_base = None

        if postgresql_base:
            postgresql_base.CursorDebugWrapper = PostgreSQLPrintCursorDebugWrapper

        if connections:
            for connection_name in connections:
                connections[connection_name].force_debug_cursor = True

        yield

        utils.CursorDebugWrapper = _CursorDebugWrapper

        if postgresql_base:
            postgresql_base.CursorDebugWrapper = _PostgreSQLCursorDebugWrapper

        if connections:
            for connection_name in connections:
                connections[connection_name].force_debug_cursor = _force_debug_cursor[
                    connection_name
                ]



# -*- coding: utf-8 -*-
"""
sqldns.py

Prints Data Source Name on stdout
"""

import sys
import warnings
from typing import List

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.core.management.color import color_style
from django.db import DEFAULT_DB_ALIAS
from django_extensions.settings import SQLITE_ENGINES, POSTGRESQL_ENGINES, MYSQL_ENGINES
from django_extensions.utils.deprecation import RemovedInNextVersionWarning


def _sqlite_name(dbhost, dbport, dbname, dbuser, dbpass):
    return dbname


def _mysql_keyvalue(dbhost, dbport, dbname, dbuser, dbpass):
    dsnstr = f'host="{dbhost}", db="{dbname}", user="{dbuser}", passwd="{dbpass}"'
    if dbport is not None:
        dsnstr += f', port="{dbport}"'
    return dsnstr


def _mysql_args(dbhost, dbport, dbname, dbuser, dbpass):
    dsnstr = f'-h "{dbhost}" -D "{dbname}" -u "{dbuser}" -p "{dbpass}"'
    if dbport is not None:
        dsnstr += f" -P {dbport}"
    return dsnstr


def _postgresql_keyvalue(dbhost, dbport, dbname, dbuser, dbpass):
    dsnstr = f"host='{dbhost}' dbname='{dbname}' user='{dbuser}' password='{dbpass}'"
    if dbport is not None:
        dsnstr += f" port='{dbport}'"
    return dsnstr


def _postgresql_kwargs(dbhost, dbport, dbname, dbuser, dbpass):
    dsnstr = (
        f"host={dbhost!r}, database={dbname!r}, user={dbuser!r}, password={dbpass!r}"
    )
    if dbport is not None:
        dsnstr += f", port={dbport!r}"
    return dsnstr


def _postgresql_pgpass(dbhost, dbport, dbname, dbuser, dbpass):
    return ":".join(str(s) for s in [dbhost, dbport, dbname, dbuser, dbpass])


def _uri(engine):
    def inner(dbhost, dbport, dbname, dbuser, dbpass):
        host = dbhost or ""
        if dbport is not None and dbport != "":
            host += f":{dbport}"
        if dbuser is not None and dbuser != "":
            user = dbuser
            if dbpass is not None and dbpass != "":
                user += f":{dbpass}"
            host = f"{user}@{host}"
        return f"{engine}://{host}/{dbname}"

    return inner


_FORMATTERS = [
    (SQLITE_ENGINES, None, _sqlite_name),
    (SQLITE_ENGINES, "filename", _sqlite_name),
    (SQLITE_ENGINES, "uri", _uri("sqlite")),
    (MYSQL_ENGINES, None, _mysql_keyvalue),
    (MYSQL_ENGINES, "keyvalue", _mysql_keyvalue),
    (MYSQL_ENGINES, "args", _mysql_args),
    (MYSQL_ENGINES, "uri", _uri("mysql")),
    (POSTGRESQL_ENGINES, None, _postgresql_keyvalue),
    (POSTGRESQL_ENGINES, "keyvalue", _postgresql_keyvalue),
    (POSTGRESQL_ENGINES, "kwargs", _postgresql_kwargs),
    (POSTGRESQL_ENGINES, "uri", _uri("postgresql")),
    (POSTGRESQL_ENGINES, "pgpass", _postgresql_pgpass),
]


class Command(BaseCommand):
    help = "Prints DSN on stdout, as specified in settings.py"
    requires_system_checks: List[str] = []
    can_import_settings = True

    def add_arguments(self, parser):
        super().add_arguments(parser)
        dbspec = parser.add_mutually_exclusive_group()
        dbspec.add_argument(
            "-R",
            "--router",
            action="store",
            dest="router",
            default=DEFAULT_DB_ALIAS,
            help=(
                "Use this router-database other then default "
                "(deprecated: use --database instead)"
            ),
        )
        dbspec.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help=(
                "Nominates a database to run command for. "
                'Defaults to the "%s" database.'
            )
            % DEFAULT_DB_ALIAS,
        )
        styles = sorted(
            set([style for _, style, _ in _FORMATTERS if style is not None])
        )
        parser.add_argument(
            "-s",
            "--style",
            action="store",
            dest="style",
            default=None,
            choices=styles + ["all"],
            help="DSN format style.",
        )
        dbspec.add_argument(
            "-a",
            "--all",
            action="store_true",
            dest="all",
            default=False,
            help="Show DSN for all database routes",
        )
        parser.add_argument(
            "-q",
            "--quiet",
            action="store_true",
            dest="quiet",
            default=False,
            help="Quiet mode only show DSN",
        )

    def handle(self, *args, **options):
        self.style = color_style()
        all_databases = options["all"]

        if all_databases:
            databases = settings.DATABASES.keys()
        else:
            databases = [options["database"]]
            if options["router"] != DEFAULT_DB_ALIAS:
                warnings.warn(
                    "--router is deprecated. You should use --database.",
                    RemovedInNextVersionWarning,
                    stacklevel=2,
                )
                databases = [options["router"]]

        for i, database in enumerate(databases):
            if i != 0:
                sys.stdout.write("\n")
            self.show_dsn(database, options)

    def show_dsn(self, database, options):
        dbinfo = settings.DATABASES.get(database)
        quiet = options["quiet"]
        dsn_style = options["style"]

        if dbinfo is None:
            raise CommandError("Unknown database %s" % database)

        engine = dbinfo.get("ENGINE")
        dbuser = dbinfo.get("USER")
        dbpass = dbinfo.get("PASSWORD")
        dbname = dbinfo.get("NAME")
        dbhost = dbinfo.get("HOST")
        dbport = dbinfo.get("PORT")
        if dbport == "":
            dbport = None

        dsn = [
            formatter(dbhost, dbport, dbname, dbuser, dbpass)
            for engines, style, formatter in _FORMATTERS
            if engine in engines
            and (dsn_style == style or dsn_style == "all" and style is not None)
        ]

        if not dsn:
            available = ", ".join(
                style
                for engines, style, _ in _FORMATTERS
                if engine in engines and style is not None
            )
            dsn = [
                self.style.ERROR(
                    f"Invalid style {dsn_style} for {engine} (available: {available})"
                    if available
                    else "Unknown database, can't generate DSN"
                )
            ]

        if not quiet:
            sys.stdout.write(
                self.style.SQL_TABLE(
                    f"DSN for database {database!r} with engine {engine!r}:\n"
                )
            )

        for output in dsn:
            sys.stdout.write(f"{output}\n")



# -*- coding: utf-8 -*-
import sys

from django.core.management.base import BaseCommand
from logging import getLogger

logger = getLogger("django.commands")


class LoggingBaseCommand(BaseCommand):
    """
    A subclass of BaseCommand that logs run time errors to `django.commands`.
    To use this, create a management command subclassing LoggingBaseCommand:

        from django_extensions.management.base import LoggingBaseCommand

        class Command(LoggingBaseCommand):
            help = 'Test error'

            def handle(self, *args, **options):
                raise Exception


    And then define a logging handler in settings.py:

        LOGGING = {
            ... # Other stuff here

            'handlers': {
                'mail_admins': {
                    'level': 'ERROR',
                    'filters': ['require_debug_false'],
                    'class': 'django.utils.log.AdminEmailHandler'
                },
            },
            'loggers': {
                'django.commands': {
                    'handlers': ['mail_admins'],
                    'level': 'ERROR',
                    'propagate': False,
                },
            }

        }

    """

    def execute(self, *args, **options):
        try:
            super().execute(*args, **options)
        except Exception as e:
            logger.error(e, exc_info=sys.exc_info(), extra={"status_code": 500})
            raise



# -*- coding: utf-8 -*-
import re

from django.template import Library
from django.utils.encoding import force_str


register = Library()
re_widont = re.compile(r"\s+(\S+\s*)$")
re_widont_html = re.compile(
    r"([^<>\s])\s+([^<>\s]+\s*)(</?(?:address|blockquote|br|dd|div|dt|fieldset|form|h[1-6]|li|noscript|p|td|th)[^>]*>|$)",
    re.IGNORECASE,
)


@register.filter
def widont(value, count=1):
    """
    Add an HTML non-breaking space between the final two words of the string to
    avoid "widowed" words.

    Examples:

    >>> print(widont('Test   me   out'))
    Test   me&nbsp;out

    >>> print("'",widont('It works with trailing spaces too  '), "'")
    ' It works with trailing spaces&nbsp;too   '

    >>> print(widont('NoEffect'))
    NoEffect
    """

    def replace(matchobj):
        return force_str("&nbsp;%s" % matchobj.group(1))

    for i in range(count):
        value = re_widont.sub(replace, force_str(value))
    return value


@register.filter
def widont_html(value):
    """
    Add an HTML non-breaking space between the final two words at the end of
    (and in sentences just outside of) block level tags to avoid "widowed"
    words.

    Examples:

    >>> print(widont_html('<h2>Here is a simple  example  </h2> <p>Single</p>'))
    <h2>Here is a simple&nbsp;example  </h2> <p>Single</p>

    >>> print(widont_html('<p>test me<br /> out</p><h2>Ok?</h2>Not in a p<p title="test me">and this</p>'))
    <p>test&nbsp;me<br /> out</p><h2>Ok?</h2>Not in a&nbsp;p<p title="test me">and&nbsp;this</p>

    >>> print(widont_html('leading text  <p>test me out</p>  trailing text'))
    leading&nbsp;text  <p>test me&nbsp;out</p>  trailing&nbsp;text
    """  # noqa: E501

    def replace(matchobj):
        return force_str("%s&nbsp;%s%s" % matchobj.groups())

    return re_widont_html.sub(replace, force_str(value))


if __name__ == "__main__":

    def _test():
        import doctest

        doctest.testmod()

    _test()



# -*- coding: utf-8 -*-
import logging
import os
import sys

from django_extensions.management.signals import post_command, pre_command


def _make_writeable(filename):
    """
    Make sure that the file is writable. Useful if our source is
    read-only.
    """
    import stat

    if sys.platform.startswith("java"):
        # On Jython there is no os.access()
        return
    if not os.access(filename, os.W_OK):
        st = os.stat(filename)
        new_permissions = stat.S_IMODE(st.st_mode) | stat.S_IWUSR
        os.chmod(filename, new_permissions)


def setup_logger(logger, stream, filename=None, fmt=None):
    """
    Set up a logger (if no handlers exist) for console output,
    and file 'tee' output if desired.
    """
    if len(logger.handlers) < 1:
        console = logging.StreamHandler(stream)
        console.setLevel(logging.DEBUG)
        console.setFormatter(logging.Formatter(fmt))
        logger.addHandler(console)
        logger.setLevel(logging.DEBUG)
        logger.propagate = False

        if filename:
            outfile = logging.FileHandler(filename)
            outfile.setLevel(logging.INFO)
            outfile.setFormatter(
                logging.Formatter("%(asctime)s " + (fmt if fmt else "%(message)s"))
            )
            logger.addHandler(outfile)


class RedirectHandler(logging.Handler):
    """Redirect logging sent to one logger (name) to another."""

    def __init__(self, name, level=logging.DEBUG):
        logging.Handler.__init__(self, level)
        self.name = name
        self.logger = logging.getLogger(name)

    def emit(self, record):
        self.logger.handle(record)


def signalcommand(func):
    """decorator for management command handle defs that sends out a pre/post signal."""

    def inner(self, *args, **kwargs):
        pre_command.send(self.__class__, args=args, kwargs=kwargs)
        try:
            ret = func(self, *args, **kwargs)
        except Exception as e:
            post_command.send(self.__class__, args=args, kwargs=kwargs, outcome=e)
            raise
        else:
            post_command.send(self.__class__, args=args, kwargs=kwargs, outcome=ret)
        return ret

    return inner


def has_ipdb():
    try:
        import ipdb  # noqa
        import IPython  # noqa

        return True
    except ImportError:
        return False



# -*- coding: utf-8 -*-
from django.apps import apps
from django.contrib.contenttypes.fields import GenericForeignKey
from django.core.management import BaseCommand
from django.db import transaction

from django_extensions.management.utils import signalcommand


def get_model_to_deduplicate():
    models = apps.get_models()
    iterator = 1
    for model in models:
        print("%s. %s" % (iterator, model.__name__))
        iterator += 1
    model_choice = int(
        input("Enter the number of the model you would like to de-duplicate:")
    )
    model_to_deduplicate = models[model_choice - 1]
    return model_to_deduplicate


def get_field_names(model):
    fields = [field.name for field in model._meta.get_fields()]
    iterator = 1
    for field in fields:
        print("%s. %s" % (iterator, field))
        iterator += 1
    validated = False
    while not validated:
        first_field = int(
            input(
                "Enter the number of the (first) field you would like to de-duplicate."
            )
        )
        if first_field in range(1, iterator):
            validated = True
        else:
            print("Invalid input. Please try again.")
    fields_to_deduplicate = [fields[first_field - 1]]

    done = False
    while not done:
        available_fields = [f for f in fields if f not in fields_to_deduplicate]
        iterator = 1
        for field in available_fields:
            print("%s. %s" % (iterator, field))
            iterator += 1
        print("C. Done adding fields.")

        validated = False
        while not validated:
            print("You are currently deduplicating on the following fields:")
            print("\n".join(fields_to_deduplicate) + "\n")

            additional_field = input("""
                Enter the number of the field you would like to de-duplicate.
                If you have entered all fields, enter C to continue.
            """)
            if additional_field == "C":
                done = True
                validated = True
            elif int(additional_field) in list(range(1, len(available_fields) + 1)):
                fields_to_deduplicate += [available_fields[int(additional_field) - 1]]
                validated = True
            else:
                print("Invalid input. Please try again.")

    return fields_to_deduplicate


def keep_first_or_last_instance():
    while True:
        first_or_last = input("""
            Do you want to keep the first or last duplicate instance?
            Enter "first" or "last" to continue.
            """)
        if first_or_last in ["first", "last"]:
            return first_or_last


def get_generic_fields():
    """Return a list of all GenericForeignKeys in all models."""
    generic_fields = []
    for model in apps.get_models():
        for field_name, field in model.__dict__.items():
            if isinstance(field, GenericForeignKey):
                generic_fields.append(field)
    return generic_fields


class Command(BaseCommand):
    help = """
        Removes duplicate model instances based on a specified
        model and field name(s).

        Makes sure that any OneToOne, ForeignKey, or ManyToMany relationships
        attached to a deleted model(s) get reattached to the remaining model.

        Based on the following:
        https://djangosnippets.org/snippets/2283/
        https://stackoverflow.com/a/41291137/2532070
        https://gist.github.com/edelvalle/01886b6f79ba0c4dce66
    """

    @signalcommand
    def handle(self, *args, **options):
        model = get_model_to_deduplicate()
        field_names = get_field_names(model)
        first_or_last = keep_first_or_last_instance()
        total_deleted_objects_count = 0
        for instance in model.objects.all():
            kwargs = {}
            for field_name in field_names:
                instance_field_value = instance.__getattribute__(field_name)
                kwargs.update({field_name: instance_field_value})
            try:
                model.objects.get(**kwargs)
            except model.MultipleObjectsReturned:
                instances = model.objects.filter(**kwargs)
                if first_or_last == "first":
                    primary_object = instances.first()
                    alias_objects = instances.exclude(pk=primary_object.pk)
                elif first_or_last == "last":
                    primary_object = instances.last()
                    alias_objects = instances.exclude(pk=primary_object.pk)

                primary_object, deleted_objects, deleted_objects_count = (
                    self.merge_model_instances(primary_object, alias_objects)
                )
                total_deleted_objects_count += deleted_objects_count

        print(
            "Successfully deleted {} model instances.".format(
                total_deleted_objects_count
            )
        )

    @transaction.atomic()
    def merge_model_instances(self, primary_object, alias_objects):
        """
        Merge several model instances into one, the `primary_object`.
        Use this function to merge model objects and migrate all of the related
        fields from the alias objects the primary object.
        """
        generic_fields = get_generic_fields()

        # get related fields
        related_fields = list(
            filter(lambda x: x.is_relation is True, primary_object._meta.get_fields())
        )

        many_to_many_fields = list(
            filter(lambda x: x.many_to_many is True, related_fields)
        )

        related_fields = list(filter(lambda x: x.many_to_many is False, related_fields))

        # Loop through all alias objects and migrate their references to the
        # primary object
        deleted_objects = []
        deleted_objects_count = 0
        for alias_object in alias_objects:
            # Migrate all foreign key references from alias object to primary
            # object.
            for many_to_many_field in many_to_many_fields:
                alias_varname = many_to_many_field.name
                related_objects = getattr(alias_object, alias_varname)
                for obj in related_objects.all():
                    try:
                        # Handle regular M2M relationships.
                        getattr(alias_object, alias_varname).remove(obj)
                        getattr(primary_object, alias_varname).add(obj)
                    except AttributeError:
                        # Handle M2M relationships with a 'through' model.
                        # This does not delete the 'through model.
                        # TODO: Allow the user to delete a duplicate 'through' model.
                        through_model = getattr(alias_object, alias_varname).through
                        kwargs = {
                            many_to_many_field.m2m_reverse_field_name(): obj,
                            many_to_many_field.m2m_field_name(): alias_object,
                        }
                        through_model_instances = through_model.objects.filter(**kwargs)
                        for instance in through_model_instances:
                            # Re-attach the through model to the primary_object
                            setattr(
                                instance,
                                many_to_many_field.m2m_field_name(),
                                primary_object,
                            )
                            instance.save()
                            # TODO: Here, try to delete duplicate instances that are
                            # disallowed by a unique_together constraint

            for related_field in related_fields:
                if related_field.one_to_many:
                    alias_varname = related_field.get_accessor_name()
                    related_objects = getattr(alias_object, alias_varname)
                    for obj in related_objects.all():
                        field_name = related_field.field.name
                        setattr(obj, field_name, primary_object)
                        obj.save()
                elif related_field.one_to_one or related_field.many_to_one:
                    alias_varname = related_field.name
                    related_object = getattr(alias_object, alias_varname)
                    primary_related_object = getattr(primary_object, alias_varname)
                    if primary_related_object is None:
                        setattr(primary_object, alias_varname, related_object)
                        primary_object.save()
                    elif related_field.one_to_one:
                        self.stdout.write(
                            "Deleted {} with id {}\n".format(
                                related_object, related_object.id
                            )
                        )
                        related_object.delete()

            for field in generic_fields:
                filter_kwargs = {}
                filter_kwargs[field.fk_field] = alias_object._get_pk_val()
                filter_kwargs[field.ct_field] = field.get_content_type(alias_object)
                related_objects = field.model.objects.filter(**filter_kwargs)
                for generic_related_object in related_objects:
                    setattr(generic_related_object, field.name, primary_object)
                    generic_related_object.save()

            if alias_object.id:
                deleted_objects += [alias_object]
                self.stdout.write(
                    "Deleted {} with id {}\n".format(alias_object, alias_object.id)
                )
                alias_object.delete()
                deleted_objects_count += 1

        return primary_object, deleted_objects, deleted_objects_count



# -*- coding: utf-8 -*-
"""
reset_db command

originally from https://www.djangosnippets.org/snippets/828/ by dnordberg
"""

import importlib.util
import os
import logging
import warnings

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS

from django_extensions.settings import SQLITE_ENGINES, POSTGRESQL_ENGINES, MYSQL_ENGINES
from django_extensions.management.mysql import parse_mysql_cnf
from django_extensions.management.utils import signalcommand
from django_extensions.utils.deprecation import RemovedInNextVersionWarning


class Command(BaseCommand):
    help = "Resets the database for this project."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            default=True,
            help="Tells Django to NOT prompt the user for input of any kind.",
        )
        parser.add_argument(
            "--no-utf8",
            action="store_true",
            dest="no_utf8_support",
            default=False,
            help="Tells Django to not create a UTF-8 charset database",
        )
        parser.add_argument(
            "-U",
            "--user",
            action="store",
            dest="user",
            default=None,
            help="Use another user for the database than defined in settings.py",
        )
        parser.add_argument(
            "-O",
            "--owner",
            action="store",
            dest="owner",
            default=None,
            help="Use another owner for creating the database than the user defined "
            "in settings or via --user",
        )
        parser.add_argument(
            "-P",
            "--password",
            action="store",
            dest="password",
            default=None,
            help="Use another password for the database than defined in settings.py",
        )
        parser.add_argument(
            "-D",
            "--dbname",
            action="store",
            dest="dbname",
            default=None,
            help="Use another database name than defined in settings.py",
        )
        parser.add_argument(
            "-R",
            "--router",
            action="store",
            dest="router",
            default=DEFAULT_DB_ALIAS,
            help="Use this router-database other than defined in settings.py",
        )
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to run command for. Defaults to the "%s".'
            % DEFAULT_DB_ALIAS,
        )
        parser.add_argument(
            "-c",
            "--close-sessions",
            action="store_true",
            dest="close_sessions",
            default=False,
            help="Close database connections before dropping database "
            "(currently works on PostgreSQL only)",
        )

    @signalcommand
    def handle(self, *args, **options):
        """
        Reset the database for this project.

        Note: Transaction wrappers are in reverse as a work around for
        autocommit, anybody know how to do this the right way?
        """
        database = options["database"]
        if options["router"] != DEFAULT_DB_ALIAS:
            warnings.warn(
                "--router is deprecated. You should use --database.",
                RemovedInNextVersionWarning,
                stacklevel=2,
            )
            database = options["router"]

        dbinfo = settings.DATABASES.get(database)
        if dbinfo is None:
            raise CommandError("Unknown database %s" % database)

        engine = dbinfo.get("ENGINE")

        user = password = database_name = database_host = database_port = ""
        if engine == "mysql":
            (user, password, database_name, database_host, database_port) = (
                parse_mysql_cnf(dbinfo)
            )

        user = options["user"] or dbinfo.get("USER") or user
        password = options["password"] or dbinfo.get("PASSWORD") or password
        owner = options["owner"] or user

        database_name = options["dbname"] or dbinfo.get("NAME") or database_name
        if database_name == "":
            raise CommandError(
                "You need to specify DATABASE_NAME in your Django settings file."
            )

        database_host = dbinfo.get("HOST") or database_host
        database_port = dbinfo.get("PORT") or database_port

        verbosity = options["verbosity"]
        if options["interactive"]:
            confirm = input(
                """
You have requested a database reset.
This will IRREVERSIBLY DESTROY
ALL data in the database "%s".
Are you sure you want to do this?

Type 'yes' to continue, or 'no' to cancel: """
                % (database_name,)
            )
        else:
            confirm = "yes"

        if confirm != "yes":
            print("Reset cancelled.")
            return

        if engine in SQLITE_ENGINES:
            try:
                logging.info("Unlinking %s database", engine)
                os.unlink(database_name)
            except OSError:
                pass
        elif engine in MYSQL_ENGINES:
            import MySQLdb as Database

            kwargs = {
                "user": user,
                "passwd": password,
            }
            if database_host.startswith("/"):
                kwargs["unix_socket"] = database_host
            else:
                kwargs["host"] = database_host

            if database_port:
                kwargs["port"] = int(database_port)

            connection = Database.connect(**kwargs)
            drop_query = "DROP DATABASE IF EXISTS `%s`" % database_name
            utf8_support = "" if options["no_utf8_support"] else "CHARACTER SET utf8"
            create_query = "CREATE DATABASE `%s` %s" % (database_name, utf8_support)
            logging.info('Executing... "%s"', drop_query)
            connection.query(drop_query)
            logging.info('Executing... "%s"', create_query)
            connection.query(create_query.strip())
        elif engine in POSTGRESQL_ENGINES:
            has_psycopg3 = importlib.util.find_spec("psycopg")
            if has_psycopg3:
                import psycopg as Database  # NOQA
            else:
                import psycopg2 as Database  # NOQA

            conn_params = {"dbname": "template1"}
            if user:
                conn_params["user"] = user
            if password:
                conn_params["password"] = password
            if database_host:
                conn_params["host"] = database_host
            if database_port:
                conn_params["port"] = database_port

            connection = Database.connect(**conn_params)
            if has_psycopg3:
                connection.autocommit = True
            else:
                connection.set_isolation_level(0)  # autocommit false
            cursor = connection.cursor()

            if options["close_sessions"]:
                close_sessions_query = (
                    """
                    SELECT pg_terminate_backend(pg_stat_activity.pid)
                    FROM pg_stat_activity
                    WHERE pg_stat_activity.datname = '%s';
                """
                    % database_name
                )
                logging.info('Executing... "%s"', close_sessions_query.strip())
                try:
                    cursor.execute(close_sessions_query)
                except Database.ProgrammingError as e:
                    logging.exception("Error: %s", str(e))

            drop_query = 'DROP DATABASE "%s";' % database_name
            logging.info('Executing... "%s"', drop_query)
            try:
                cursor.execute(drop_query)
            except Database.ProgrammingError as e:
                logging.exception("Error: %s", str(e))

            create_query = 'CREATE DATABASE "%s"' % database_name
            if owner:
                create_query += ' WITH OWNER = "%s" ' % owner
            create_query += " ENCODING = 'UTF8'"

            if settings.DEFAULT_TABLESPACE:
                create_query += " TABLESPACE = %s;" % settings.DEFAULT_TABLESPACE
            else:
                create_query += ";"

            logging.info('Executing... "%s"', create_query)
            cursor.execute(create_query)
        else:
            raise CommandError("Unknown database engine %s" % engine)

        if verbosity >= 2 or options["interactive"]:
            print("Reset successful.")



# -*- coding: utf-8 -*-
#
# Autocomplete feature for admin panel
#
import operator
from functools import update_wrapper, reduce
from typing import Tuple, Dict, Callable  # NOQA

from django.apps import apps
from django.http import HttpResponse, HttpResponseNotFound
from django.conf import settings
from django.db import models
from django.db.models.query import QuerySet
from django.utils.encoding import smart_str
from django.utils.translation import gettext as _
from django.utils.text import get_text_list
from django.contrib import admin

from django_extensions.admin.widgets import ForeignKeySearchInput


class ForeignKeyAutocompleteAdminMixin:
    """
    Admin class for models using the autocomplete feature.

    There are two additional fields:
       - related_search_fields: defines fields of managed model that
         have to be represented by autocomplete input, together with
         a list of target model fields that are searched for
         input string, e.g.:

         related_search_fields = {
            'author': ('first_name', 'email'),
         }

       - related_string_functions: contains optional functions which
         take target model instance as only argument and return string
         representation. By default __unicode__() method of target
         object is used.

    And also an optional additional field to set the limit on the
    results returned by the autocomplete query. You can set this integer
    value in your settings file using FOREIGNKEY_AUTOCOMPLETE_LIMIT or
    you can set this per ForeignKeyAutocompleteAdmin basis. If any value
    is set the results will not be limited.
    """

    related_search_fields = {}  # type: Dict[str, Tuple[str]]
    related_string_functions = {}  # type: Dict[str, Callable]
    autocomplete_limit = getattr(settings, "FOREIGNKEY_AUTOCOMPLETE_LIMIT", None)

    def get_urls(self):
        from django.urls import path

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)

            return update_wrapper(wrapper, view)

        return [
            path(
                "foreignkey_autocomplete/",
                wrap(self.foreignkey_autocomplete),
                name="%s_%s_autocomplete"
                % (self.model._meta.app_label, self.model._meta.model_name),
            )
        ] + super().get_urls()

    def foreignkey_autocomplete(self, request):
        """
        Search in the fields of the given related model and returns the
        result as a simple string to be used by the jQuery Autocomplete plugin
        """
        query = request.GET.get("q", None)
        app_label = request.GET.get("app_label", None)
        model_name = request.GET.get("model_name", None)
        search_fields = request.GET.get("search_fields", None)
        object_pk = request.GET.get("object_pk", None)

        try:
            to_string_function = self.related_string_functions[model_name]
        except KeyError:
            to_string_function = lambda x: x.__str__()

        if search_fields and app_label and model_name and (query or object_pk):

            def construct_search(field_name):
                # use different lookup methods depending on the notation
                if field_name.startswith("^"):
                    return "%s__istartswith" % field_name[1:]
                elif field_name.startswith("="):
                    return "%s__iexact" % field_name[1:]
                elif field_name.startswith("@"):
                    return "%s__search" % field_name[1:]
                else:
                    return "%s__icontains" % field_name

            model = apps.get_model(app_label, model_name)

            queryset = model._default_manager.all()
            data = ""
            if query:
                for bit in query.split():
                    or_queries = [
                        models.Q(
                            **{construct_search(smart_str(field_name)): smart_str(bit)}
                        )
                        for field_name in search_fields.split(",")
                    ]
                    other_qs = QuerySet(model)
                    other_qs.query.select_related = queryset.query.select_related
                    other_qs = other_qs.filter(reduce(operator.or_, or_queries))
                    queryset = queryset & other_qs

                additional_filter = self.get_related_filter(model, request)
                if additional_filter:
                    queryset = queryset.filter(additional_filter)

                if self.autocomplete_limit:
                    queryset = queryset[: self.autocomplete_limit]

                data = "".join(
                    [str("%s|%s\n") % (to_string_function(f), f.pk) for f in queryset]
                )
            elif object_pk:
                try:
                    obj = queryset.get(pk=object_pk)
                except Exception:  # FIXME: use stricter exception checking
                    pass
                else:
                    data = to_string_function(obj)
            return HttpResponse(data, content_type="text/plain")
        return HttpResponseNotFound()

    def get_related_filter(self, model, request):
        """
        Given a model class and current request return an optional Q object
        that should be applied as an additional filter for autocomplete query.
        If no additional filtering is needed, this method should return
        None.
        """
        return None

    def get_help_text(self, field_name, model_name):
        searchable_fields = self.related_search_fields.get(field_name, None)
        if searchable_fields:
            help_kwargs = {
                "model_name": model_name,
                "field_list": get_text_list(searchable_fields, _("and")),
            }
            return (
                _(
                    "Use the left field to do %(model_name)s lookups "
                    "in the fields %(field_list)s."
                )
                % help_kwargs
            )
        return ""

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        """
        Override the default widget for Foreignkey fields if they are
        specified in the related_search_fields class attribute.
        """
        if (
            isinstance(db_field, models.ForeignKey)
            and db_field.name in self.related_search_fields
        ):
            help_text = self.get_help_text(
                db_field.name, db_field.remote_field.model._meta.object_name
            )
            if kwargs.get("help_text"):
                help_text = str("%s %s") % (kwargs["help_text"], help_text)
            kwargs["widget"] = ForeignKeySearchInput(
                db_field.remote_field, self.related_search_fields[db_field.name]
            )
            kwargs["help_text"] = help_text
        return super().formfield_for_dbfield(db_field, request, **kwargs)


class ForeignKeyAutocompleteAdmin(ForeignKeyAutocompleteAdminMixin, admin.ModelAdmin):
    pass


class ForeignKeyAutocompleteTabularInline(
    ForeignKeyAutocompleteAdminMixin, admin.TabularInline
):
    pass


class ForeignKeyAutocompleteStackedInline(
    ForeignKeyAutocompleteAdminMixin, admin.StackedInline
):
    pass
